# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Watchpoints."""
from abc import ABC
from enum import Enum

from mindinsight.debugger.api.debugger_tensor import DebuggerTensor
from mindinsight.debugger.common.exceptions.exceptions import DebuggerParamValueError, \
    DebuggerParamTypeError
from mindinsight.debugger.common.utils import validate_type
from mindinsight.debugger.conditionmgr.condition import ParamNameEnum


class ConditionBase(ABC):
    """
    Base class for watch conditions.

    .. warning::
        All APIs in this class are experimental prototypes that are subject to
        change or deletion.

    Note:
        - If multiple checking parameters are specified for one condition instance,
          a `WatchpointHit` happens for the parameters that the tensor triggered for the watchpoint.

    Supported Platforms:
        ``Ascend`` ``GPU``

    Examples:
            >>> from mindinsight.debugger import DumpAnalyzer
            >>> from mindinsight.debugger import (TensorTooLargeCondition,
            ...                                   Watchpoint)
            >>>
            >>> def test_condition_base():
            ...     my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
            ...     tensors = my_run.select_tensors(query_string="Conv2D-op13")
            ...     watchpoint = Watchpoint(tensors=tensors,
            ...                             condition=TensorTooLargeCondition(abs_mean_gt=0.0, max_gt=0.0))
            ...     hit = list(my_run.check_watchpoints(watchpoints=[watchpoint]))[0]
            ...     # print(hit.get_hit_detail())
            ...     # the print result is as follows
            ...     # The setting for watchpoint is abs_mean_gt = 0.0, max_gt = 0.0.
            ...     # The actual value of the tensor is abs_mean_gt = 0.06592023578438996, max_gt = 0.449951171875.
            ...     watchpoint = Watchpoint(tensors=tensors,
            ...                             condition=TensorTooLargeCondition(abs_mean_gt=0.0, max_gt=1.0))
            ...     # the check_watchpoints function start a new process needs to be called through the main entry
            ...     hit = list(my_run.check_watchpoints(watchpoints=[watchpoint]))[0]
            ...     # print(hit.get_hit_detail())
            ...     # the print result is as follows
            ...     # The setting for watchpoint is abs_mean_gt = 0.0.
            ...     # The actual value of the tensor is abs_mean_gt = 0.06592023578438996.
            ...
            >>> if __name__ == "__main__":
            ...     test_condition_base()
            ...
    """

    @property
    def name(self):
        """
        Get the id for watch condition.

        Returns:
            str, the name of the watch condition.
        """
        raise NotImplementedError

    @property
    def condition_id(self):
        """
        Get the name for the watch condition Id.

        Returns:
            int, the id of the watch condition.
        """
        raise NotImplementedError

    @property
    def param_dict(self):
        """
        Get the parameters list.

        Returns:
            dict, the parameter dict of the watch condition.
        """
        return {}

    def __str__(self):
        return str(self.param_dict)


class WatchpointHit(ABC):
    """
    Watchpoint hit.

    .. warning::
        All APIs in this class are experimental prototypes that are subject to
        change or deletion.

    Note:
        - This class is not meant to be instantiated by user.
        - The instances of this class is immutable.

    Supported Platforms:
        ``Ascend`` ``GPU``

    Examples:
        >>> from mindinsight.debugger import DumpAnalyzer
        >>> from mindinsight.debugger import TensorTooLargeCondition, Watchpoint
        >>>
        >>> def test_watch_point_hit():
        ...     my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
        ...     tensor_list = my_run.select_tensors(
        ...                                         query_string="Conv",
        ...                                         use_regex=True,
        ...                                         iterations=[0],
        ...                                         ranks=[0],
        ...                                         slots=[0]
        ...                                         )
        ...     watchpoint = Watchpoint(tensors=tensor_list,
        ...                             condition=TensorTooLargeCondition(abs_mean_gt=0.0))
        ...     # the check_watchpoints function start a new process needs to be called through the main entry
        ...     hits = my_run.check_watchpoints(watchpoints=[watchpoint])
        ...     hit = list(hits)[0]
        ...     # print(str(hit))
        ...     # the print result is as follows
        ...     # Watchpoint TensorTooLarge triggered on tensor:
        ...     # rank: 0
        ...     # graph_name: kernel_graph_0
        ...     # node_name: Default/network-WithLossCell/_backbone-AlexNet/conv1-Conv2d/Cast-op7
        ...     # slot: 0
        ...     # iteration: 0
        ...     # Threshold: {'abs_mean_gt': 0.0}
        ...     # Hit detail: The setting for watchpoint is abs_mean_gt = 0.0.
        ...     # The actual value of the tensor is abs_mean_gt = 0.007956420533235841.
        ...     # print(hit.error_code)
        ...     # the print result is as follows
        ...     # 0
        ...     # print(hit.tensor)
        ...     # the print result is as follows
        ...     # rank: 0
        ...     # graph_name: kernel_graph_0
        ...     # node_name: Default/network-WithLossCell/_backbone-AlexNet/conv1-Conv2d/Cast-op7
        ...     # slot: 0
        ...     # iteration: 0
        ...     # print(hit.get_hit_detail())
        ...     # the print result is as follows
        ...     # The setting for watchpoint is abs_mean_gt = 0.0.
        ...     # The actual value of the tensor is abs_mean_gt = 0.007956420533235841.
        ...
        >>> if __name__ == "__main__":
        ...     test_watch_point_hit()
        ...
    """

    @property
    def error_code(self):
        """
        Get the error code when checking the watchpoint if there is error.

        Returns:
            int, the error number.
        """
        raise NotImplementedError

    @property
    def error_msg(self):
        """
        Get the error msg when checking the watchpoint if there is error.

        Returns:
            list[str], the error message list.
        """
        raise NotImplementedError

    @property
    def tensor(self) -> DebuggerTensor:
        """
        Get the tensor for this watchpoint hit.

        Returns:
            DebuggerTensor, the triggered tensor.
        """
        raise NotImplementedError

    def get_threshold(self):
        """
        Get the condition set by user.

        Returns:
            ConditionBase, the condition with user threshold.
        """
        raise NotImplementedError

    def get_hit_detail(self):
        """
        Get the corresponding watch condition，including the actual values.
        For example, if the corresponding watch condition is `TensorTooLargeCondition(max_gt=None)` ,
        watching whether the max value of the tensor greater than 0, the `get_hit_detail` return
        a `TensorTooLargeCondition` object including the max value of the tensor.
        If error_code is not zero, None will be returned.

        Returns:
            Union[ConditionBase, None], the condition with hit detail, If error_code is not zero,
            None will be returned.
        """
        raise NotImplementedError


class WatchpointHitImpl(WatchpointHit):
    """
    Watchpoint hit.

    Args:
        tensor (DebuggerTensor): The tensor which hits the watchpoint.
        condition (ConditionBase): The ConditionBase object initialized with
            user setting value.
        hit_detail (ConditionBase): The ConditionBase object
            initialized with actual value of the Tensor.
        error_code (int): The code describing error.
    """

    def __init__(self,
                 tensor: DebuggerTensor,
                 condition: ConditionBase,
                 hit_detail: ConditionBase,
                 error_code):
        self._tensor = tensor
        self._condition = condition
        self._error_code = error_code
        self._hit_detail = hit_detail

    @property
    def error_code(self):
        """
        Get the error code when checking the watchpoint if there is error.

        Returns:
            int, the error number.
        """
        return self._error_code

    @property
    def error_msg(self):
        """
        Get the error msg when checking the watchpoint if there is error.

        Returns:
            list[str], the error message list.
        """
        error_code = self._error_code
        all_error_list = [
            "Tensor contains NaN.",
            "A tensor contains +/-INF.",
            "The previous step value cannot be found.",
            "The tensor size exceeds the memory limit.",
            "Graph history file is not available.",
            "Tensor has no value."
        ]
        error_list = []
        for i, error_str in enumerate(all_error_list):
            error = (error_code >> i) & 1
            if error == 1:
                error_list.append(error_str)

        return error_list

    @property
    def tensor(self) -> DebuggerTensor:
        """Get the tensor for this watchpoint hit."""
        return self._tensor

    def get_threshold(self):
        """Get the threshold set by user."""
        return self._condition

    def get_hit_detail(self):
        """
        Get the actual values for the thresholds in the watchpoint.
        If error_code is not zero or None, None will be returned.
        """
        if self._error_code:
            return None
        return self._hit_detail

    def __str__(self):
        if self._error_code:
            msg = f"Watchpoint {self._condition.name} check failed on tensor:\n" \
                  f"{str(self.tensor)}" \
                  f"Threshold: {self.get_threshold()}\n" \
                  f"Error detail: {self.error_msg}"
            return msg
        msg = f"Watchpoint {self._condition.name} triggered on tensor:\n" \
              f"{str(self.tensor)}" \
              f"Threshold: {self.get_threshold()}\n" \
              f"Hit detail: {str(self._hit_detail)}"
        return msg


class HitDetail(ConditionBase):
    """Hit Detail."""

    def __init__(self, param_list, condition):
        self._param_list = param_list
        self._condition = condition

    @property
    def name(self):
        """Get the name for the condition."""
        return self._condition.name

    @property
    def condition_id(self):
        """Get the name for the condition Id."""
        return self._condition.condition_id

    @property
    def param_dict(self):
        """Get the parameters list."""
        return self._param_list

    def __str__(self):
        show_actual_value = bool(self._condition.param_dict)
        if self._condition.condition_id == WatchpointConditionId.UNCHANGED_TENSOR.value:
            show_actual_value = False
        # list of the parameters with disabled = False and hit = 1
        hit_param_list = []
        for param in self._param_list:
            if not param.disabled and param.hit:
                hit_param_list.append(param)

        result = ""
        param_size = len(hit_param_list)
        if show_actual_value and hit_param_list:
            setting_detail = "The setting for watchpoint is "
            value_detail = " The actual value of the tensor is "
            for idx, param in enumerate(hit_param_list):
                setting_detail += f"{param.name} = {param.value}"
                value_detail += f"{param.name} = {param.actual_value}"
                if idx == param_size - 1:
                    setting_detail += "."
                    value_detail += "."
                else:
                    setting_detail += ", "
                    value_detail += ", "
            result = setting_detail + value_detail

        if not result:
            result = "None."
        return result


class TensorTooLargeCondition(ConditionBase):
    """
    Watch contion for tensor value too large. At least one parameter should be specified.

    If multiple checking parameters are specified, a `WatchpointHit` happens for the parameters
    that the tensor triggered for the watchpoint.

    .. warning::
        All APIs in this class are experimental prototypes that are subject to
        change or deletion.

    Args:
        abs_mean_gt (float, optional): The threshold for mean of the absolute
            value of the tensor. When the actual value was greater than this
            threshold, this checking condition would be satisfied. Default: ``None``.
        max_gt (float, optional): The threshold for maximum of the tensor. When
            the actual value was greater than this threshold, this checking
            condition would be satisfied. Default: ``None``.
        min_gt (float, optional): The threshold for minimum of the tensor. When
            the actual value was greater than this threshold, this checking
            condition would be satisfied. Default: ``None``.
        mean_gt (float, optional): The threshold for mean of the tensor. When
            the actual value was greater than this threshold, this checking
            condition would be satisfied. Default: ``None``.

    Examples:
        >>> from mindinsight.debugger import TensorTooLargeCondition
        >>> my_condition = TensorTooLargeCondition(abs_mean_gt=0.0)
        >>> print(my_condition.name)
        TensorTooLarge
    """

    def __init__(self,
                 abs_mean_gt=None, max_gt=None, min_gt=None, mean_gt=None):
        self._abs_mean_gt = abs_mean_gt
        self._max_gt = max_gt
        self._min_gt = min_gt
        self._mean_gt = mean_gt
        self._param_dict = self._get_param_dict()

    @property
    def name(self):
        return "TensorTooLarge"

    @property
    def condition_id(self):
        return WatchpointConditionId.TENSOR_TOO_LARGE.value

    @property
    def param_dict(self):
        return self._param_dict

    def _get_param_dict(self):
        """Get normalized param dict."""
        param_dict = {}
        if self._abs_mean_gt is not None:
            validate_type(self._abs_mean_gt, 'abs_mean_gt', [int, float], 'float')
            param_dict[ParamNameEnum.ABS_MEAN_GT.value] = float(self._abs_mean_gt)
        if self._max_gt is not None:
            validate_type(self._max_gt, 'max_gt', [int, float], 'float')
            param_dict[ParamNameEnum.MAX_GT.value] = float(self._max_gt)
        if self._min_gt is not None:
            validate_type(self._min_gt, 'min_gt', [int, float], 'float')
            param_dict[ParamNameEnum.MIN_GT.value] = float(self._min_gt)
        if self._mean_gt is not None:
            validate_type(self._mean_gt, 'mean_gt', [int, float], 'float')
            param_dict[ParamNameEnum.MEAN_GT.value] = float(self._mean_gt)
        if not param_dict:
            msg = "Please specify at least one of the parameters for TensorTooLargeCondition."
            raise DebuggerParamValueError(msg)
        return param_dict

    @property
    def param_names(self):
        """
        Return the list of parameter names.

        Returns:
            list[str], the parameter names.
        """
        names = [
            ParamNameEnum.ABS_MEAN_GT.value,
            ParamNameEnum.MAX_GT.value,
            ParamNameEnum.MIN_GT.value,
            ParamNameEnum.MEAN_GT.value
        ]
        return names


class TensorTooSmallCondition(ConditionBase):
    """
    Watch contion for tensor value too small. At least one parameter should be specified.

    If multiple checking parameters are specified, a `WatchpointHit` happens for the parameters
    that the tensor triggered for the watchpoint.

    .. warning::
        All APIs in this class are experimental prototypes that are subject to
        change or deletion.

    Args:
        abs_mean_lt (float, optional): The threshold for mean of the absolute
            value of the tensor. When the actual value was less than this
            threshold, this checking condition would be satisfied. Default: ``None``.
        max_lt (float, optional): The threshold for maximum of the tensor. When
            the actual value was less than this threshold, this checking
            condition would be satisfied. Default: ``None``.
        min_lt (float, optional): The threshold for minimum of the tensor. When
            the actual value was less than this threshold, this checking
            condition would be satisfied. Default: ``None``.
        mean_lt (float, optional): The threshold for mean of the tensor. When
            the actual value was less than this threshold, this checking
            condition would be satisfied. Default: ``None``.

    Examples:
        >>> from mindinsight.debugger import TensorTooSmallCondition
        >>> my_condition = TensorTooSmallCondition(abs_mean_lt=0.2)
        >>> print(my_condition.name)
        TensorTooSmall
    """

    def __init__(self,
                 abs_mean_lt=None, max_lt=None, min_lt=None, mean_lt=None):
        self._abs_mean_lt = abs_mean_lt
        self._max_lt = max_lt
        self._min_lt = min_lt
        self._mean_lt = mean_lt
        self._param_dict = self._get_param_dict()

    @property
    def name(self):
        return "TensorTooSmall"

    @property
    def condition_id(self):
        return WatchpointConditionId.TENSOR_TOO_SMALL.value

    @property
    def param_dict(self):
        return self._param_dict

    def _get_param_dict(self):
        """Get normalized param dict."""
        param_dict = {}
        if self._abs_mean_lt is not None:
            validate_type(self._abs_mean_lt, 'abs_mean_lt', [int, float], 'float')
            param_dict[ParamNameEnum.ABS_MEAN_LT.value] = float(self._abs_mean_lt)
        if self._max_lt is not None:
            validate_type(self._max_lt, 'max_lt', [int, float], 'float')
            param_dict[ParamNameEnum.MAX_LT.value] = float(self._max_lt)
        if self._min_lt is not None:
            validate_type(self._min_lt, 'min_lt', [int, float], 'float')
            param_dict[ParamNameEnum.MIN_LT.value] = float(self._min_lt)
        if self._mean_lt is not None:
            validate_type(self._mean_lt, 'mean_lt', [int, float], 'float')
            param_dict[ParamNameEnum.MEAN_LT.value] = float(self._mean_lt)
        if not param_dict:
            msg = "Please specify at least one of the parameters for TensorTooSmallCondition."
            raise DebuggerParamValueError(msg)
        return param_dict

    @property
    def param_names(self):
        """
        Return the list of parameter names.

        Returns:
            list[str], the parameter names.
        """
        names = [
            ParamNameEnum.ABS_MEAN_LT.value,
            ParamNameEnum.MAX_LT.value,
            ParamNameEnum.MIN_LT.value,
            ParamNameEnum.MEAN_LT.value
        ]
        return names


class TensorRangeCondition(ConditionBase):
    """
    Watch condition for tensor value range.

    Set a threshold to check the tensor value range. There are four options:
    `range_percentage_lt` , `range_percentage_gt` , `max_min_lt` and `max_min_gt` .
    At least one of the four options should be specified.
    If the threshold is set to one of the first two options,
    both `range_start_inclusive` and `range_end_inclusive` must be set.
    If multiple checking parameters are specified, a `WatchpointHit` happens for the parameters
    that the tensor triggered for the watchpoint.

    .. warning::
        All APIs in this class are experimental prototypes that are subject to
        change or deletion.

    Args:
        range_start_inclusive (float, optional): The start of the specified range. Default: ``None``.
        range_end_inclusive (float, optional): The end of the specified range. Default: ``None``.
        range_percentage_lt (float, optional): The threshold for the percentage of the tensor
            in the range `[range_start_inclusive, range_end_inclusive]` . The checking condition
            will be satisfied when the percentage of the tensor in the specified range is less than this value.
            Default: ``None``.
        range_percentage_gt (float, optional): The threshold for the percentage of the tensor
            in the range `[range_start_inclusive, range_end_inclusive]` . The checking condition
            will be satisfied when the percentage of the tensor in the specified range is greater than this value.
            Default: ``None``.
        max_min_lt (float, optional): Lowwer threshold for the difference
            between the maximum and minimum values of a tensor. Default: ``None``.
        max_min_gt (float, optional): Upper threshold for the difference
            between the maximum and minimum values of a tensor. Default: ``None``.

    Examples:
        >>> from mindinsight.debugger import TensorRangeCondition
        >>> my_condition = TensorRangeCondition(max_min_gt=0.05)
        >>> print(my_condition.name)
        TensorRange
    """

    def __init__(self,
                 range_start_inclusive=None, range_end_inclusive=None, range_percentage_lt=None,
                 range_percentage_gt=None, max_min_lt=None, max_min_gt=None):
        self._range_start_inclusive = range_start_inclusive
        self._range_end_inclusive = range_end_inclusive
        self._range_percentage_lt = range_percentage_lt
        self._range_percentage_gt = range_percentage_gt
        self._max_min_lt = max_min_lt
        self._max_min_gt = max_min_gt
        self._param_dict = self._get_param_dict()

    @property
    def name(self):
        return "TensorRange"

    @property
    def condition_id(self):
        return WatchpointConditionId.TENSOR_RANGE.value

    @property
    def param_dict(self):
        return self._param_dict

    def _get_param_dict(self):
        """Get normalized param dict."""
        param_dict = {}
        if self._range_start_inclusive is not None:
            validate_type(self._range_start_inclusive, 'range_start_inclusive', [int, float], 'float')
            param_dict[ParamNameEnum.RANGE_START_INCLUSIVE.value] = float(self._range_start_inclusive)
        if self._range_end_inclusive is not None:
            validate_type(self._range_end_inclusive, 'range_end_inclusive', [int, float], 'float')
            param_dict[ParamNameEnum.RANGE_END_INCLUSIVE.value] = float(self._range_end_inclusive)
        if self._range_percentage_lt is not None:
            validate_type(self._range_percentage_lt, 'range_percentage_lt', [int, float], 'float')
            param_dict[ParamNameEnum.RANGE_PERCENTAGE_LT.value] = float(self._range_percentage_lt)
        if self._range_percentage_gt is not None:
            validate_type(self._range_percentage_gt, 'range_range_percentage_gt', [int, float], 'float')
            param_dict[ParamNameEnum.RANGE_PERCENTAGE_GT.value] = float(self._range_percentage_gt)
        if self._max_min_lt is not None:
            validate_type(self._max_min_lt, 'max_min_lt', [int, float], 'float')
            param_dict[ParamNameEnum.MAX_MIN_LT.value] = float(self._max_min_lt)
        if self._max_min_gt is not None:
            validate_type(self._max_min_gt, 'max_min_gt', [int, float], 'float')
            param_dict[ParamNameEnum.MAX_MIN_GT.value] = float(self._max_min_gt)
        if not self._has_threshold_param(param_dict):
            msg = "Please specify at least one of the parameters " \
                  "[range_percentage_lt, range_percentage_gt, max_min_lt, max_min_gt] " \
                  "for TensorRangeCondition."
            raise DebuggerParamValueError(msg)
        # check supported parameter
        if (ParamNameEnum.RANGE_PERCENTAGE_LT.value in param_dict.keys() or
                ParamNameEnum.RANGE_PERCENTAGE_GT.value in param_dict.keys()):
            if (ParamNameEnum.RANGE_START_INCLUSIVE.value not in param_dict.keys() or
                    ParamNameEnum.RANGE_END_INCLUSIVE.value not in param_dict.keys()):
                msg = ("Please specify both range_start_inclusive and "
                       "range_end_inclusive parameters for TensorRangeCondition.")
                raise DebuggerParamValueError(msg)
        return param_dict

    @staticmethod
    def _has_threshold_param(param_dict):
        """Check if threshold parameter is set."""
        threshold_param_name = [
            ParamNameEnum.RANGE_PERCENTAGE_LT.value,
            ParamNameEnum.RANGE_PERCENTAGE_GT.value,
            ParamNameEnum.MAX_MIN_LT.value,
            ParamNameEnum.MAX_MIN_GT.value
        ]
        for param_name in threshold_param_name:
            if param_name in param_dict:
                return True
        return False

    @property
    def param_names(self):
        """
        Return the list of parameter names.

        Returns:
            list[str], the parameter names.
        """
        names = [
            ParamNameEnum.RANGE_START_INCLUSIVE.value,
            ParamNameEnum.RANGE_END_INCLUSIVE.value,
            ParamNameEnum.RANGE_PERCENTAGE_LT.value,
            ParamNameEnum.RANGE_PERCENTAGE_GT.value,
            ParamNameEnum.MAX_MIN_LT.value,
            ParamNameEnum.MAX_MIN_GT.value
        ]
        return names


class TensorOverflowCondition(ConditionBase):
    """
    Watch condition for tensor overflow.

    Tensor overflow whatchpoint checks for `inf` and `nan` tensors.

    .. warning::
        All APIs in this class are experimental prototypes that are subject to
        change or deletion.

    Examples:
        >>> from mindinsight.debugger import TensorOverflowCondition
        >>> my_condition = TensorOverflowCondition()
        >>> print(my_condition.name)
        TensorOverflow
    """

    def __init__(self):
        pass

    @property
    def name(self):
        return "TensorOverflow"

    @property
    def condition_id(self):
        return WatchpointConditionId.TENSOR_OVERFLOW.value

    @property
    def param_names(self):
        """
        Return the list of parameter names.

        Returns:
            list[str], the parameter names.
        """
        return []


class OperatorOverflowCondition(ConditionBase):
    """
    Operator overflow watch condition.

    Operator overflow whatchpoint checks whether overflow occurs during operator computation.
    Only Ascend AI processor is supported.

    .. warning::
        All APIs in this class are experimental prototypes that are subject to
        change or deletion.

    Examples:
        >>> from mindinsight.debugger import OperatorOverflowCondition
        >>> my_condition = OperatorOverflowCondition()
        >>> print(my_condition.name)
        OperatorOverflow
    """

    def __init__(self):
        pass

    @property
    def name(self):
        return "OperatorOverflow"

    @property
    def condition_id(self):
        return WatchpointConditionId.OPERATOR_OVERFLOW.value

    @property
    def param_names(self):
        """
        Return the list of parameter names.

        Returns:
            list[str], the parameter names.
        """
        return []


class TensorAllZeroCondition(ConditionBase):
    """
    Watch condition for tensor value is all zero .

    .. warning::
        All APIs in this class are experimental prototypes that are subject to
        change or deletion.

    Args:
        zero_percentage_ge (float): The threshold to check if the percentage of
            zero tensor values are greater than this value.

    Examples:
        >>> from mindinsight.debugger import TensorAllZeroCondition
        >>> my_condition = TensorAllZeroCondition(zero_percentage_ge=0.0)
        >>> print(my_condition.name)
        TensorAllZero
    """

    def __init__(self, zero_percentage_ge):
        validate_type(zero_percentage_ge, 'zero_percentage_ge', [int, float], 'float')
        self._zero_percentage_ge = float(zero_percentage_ge)

    @property
    def name(self):
        return "TensorAllZero"

    @property
    def condition_id(self):
        return WatchpointConditionId.TENSOR_ALL_ZERO.value

    @property
    def param_dict(self):
        param_dict = {ParamNameEnum.ZERO_PERCENTAGE_GE.value: self._zero_percentage_ge}
        return param_dict

    @property
    def param_names(self):
        """
        Return the list of parameter names.

        Returns:
            list[str], the parameter names.
        """
        return [ParamNameEnum.ZERO_PERCENTAGE_GE.value]


class TensorUnchangedCondition(ConditionBase):
    r"""
    Watch condition for tensor value unchanged.

    Check allclose function on previous and current tensor. Only when every element in tensor
    satisfies the equation :math:`|element\_in\_current\_tensor - element\_in\_previous\_tensor|
    \leq atol + rtol\times |previous\_tensor|` , this watchpoint will be hit.

    .. warning::
        All APIs in this class are experimental prototypes that are subject to
        change or deletion.

    Args:
        rtol (float, optional): The relative tolerance parameter. Default: ``1e-5``.
        atol (float, optional): The absolute tolerance parameter. Default: ``1e-8``.

    Examples:
        >>> from mindinsight.debugger import TensorUnchangedCondition
        >>> my_condition = TensorUnchangedCondition(rtol=1000.0)
        >>> print(my_condition.name)
        TensorUnchanged
    """

    def __init__(self, rtol=1e-5, atol=1e-8):
        validate_type(rtol, 'rtol', [float, int], 'float or int')
        validate_type(atol, 'atol', [float, int], 'float or int')
        self._rtol = float(rtol)
        self._atol = float(atol)

    @property
    def name(self):
        return "TensorUnchanged"

    @property
    def condition_id(self):
        return WatchpointConditionId.UNCHANGED_TENSOR.value

    @property
    def param_dict(self):
        param_dict = {
            ParamNameEnum.RTOL.value: self._rtol,
            ParamNameEnum.ATOL.value: self._atol}
        return param_dict

    @property
    def param_names(self):
        """
        Return the list of parameter names.

        Returns:
            list[str], the parameter names.
        """
        names = [
            ParamNameEnum.RTOL.value,
            ParamNameEnum.ATOL.value,
            ParamNameEnum.EQUAL_NAN.value
        ]
        return names


class TensorChangeBelowThresholdCondition(ConditionBase):
    r"""
    Watch condition for tensor changing below threshold.

    When the tensor changing satisfies equation :math:`\frac {abs\_mean(current\_tensor
    - previous\_tensor)} {abs\_mean(previous\_tensor)} + epsilon < mean\_update\_ratio\_lt` ,
    the watchpoint would be hit.

    .. warning::
        All APIs in this class are experimental prototypes that are subject to
        change or deletion.

    Args:
        abs_mean_update_ratio_lt (float): The threshold value for mean update ration.
            If the mean update ratio is less that this value the watchpoint will be triggered.
        epsilon (float, optional): Epsilon value. Default: ``1e-9``.

    Examples:
        >>> from mindinsight.debugger import TensorChangeBelowThresholdCondition
        >>> my_condition = TensorChangeBelowThresholdCondition(abs_mean_update_ratio_lt=2.0)
        >>> print(my_condition.name)
        TensorChangeBelowThreshold
    """

    def __init__(self, abs_mean_update_ratio_lt, epsilon=1e-9):
        validate_type(abs_mean_update_ratio_lt, 'abs_mean_update_ratio_lt', [float, int], 'float')
        validate_type(epsilon, 'epsilon', [float, int], 'float')
        self._abs_mean_update_ratio_lt = float(abs_mean_update_ratio_lt)
        self._epsilon = float(epsilon)

    @property
    def name(self):
        return "TensorChangeBelowThreshold"

    @property
    def condition_id(self):
        return WatchpointConditionId.TENSOR_CHANGE_TOO_SMALL.value

    @property
    def param_dict(self):
        param_dict = {
            ParamNameEnum.ABS_MEAN_UPDATE_RATIO_LT.value: self._abs_mean_update_ratio_lt,
            ParamNameEnum.EPSILON.value: self._epsilon
        }
        return param_dict

    @property
    def param_names(self):
        """
        Return the list of parameter names.

        Returns:
            list[str], the parameter names.
        """
        names = [
            ParamNameEnum.ABS_MEAN_UPDATE_RATIO_LT.value,
            ParamNameEnum.EPSILON.value
        ]
        return names


class TensorChangeAboveThresholdCondition(ConditionBase):
    r"""
    Watch condition for tensor changing above threshold.

    When the tensor changing satisfies equation :math:`\frac {abs\_mean(current\_tensor -
    previous\_tensor)} {abs\_mean(previous\_tensor)} + epsilon > mean\_update\_ratio\_lt` ,
    the watchpoint would be hit.

    .. warning::
        All APIs in this class are experimental prototypes that are subject to
        change or deletion.

    Args:
        abs_mean_update_ratio_gt (float): The threshold value for mean update ratio,
            if the mean update ratio is greater than this value the watchpoint will be triggered.
        epsilon (float, optional): Epsilon value. Default: ``1e-9``.

    Examples:
        >>> from mindinsight.debugger import TensorChangeAboveThresholdCondition
        >>> my_condition = TensorChangeAboveThresholdCondition(abs_mean_update_ratio_gt=0.0)
        >>> print(my_condition.name)
        TensorChangeAboveThreshold
    """

    def __init__(self, abs_mean_update_ratio_gt, epsilon=1e-9):
        validate_type(abs_mean_update_ratio_gt, 'abs_mean_update_ratio_gt', [float, int], 'float')
        validate_type(epsilon, 'epsilon', [float, int], 'float')
        self._abs_mean_update_ratio_gt = float(abs_mean_update_ratio_gt)
        self._epsilon = float(epsilon)

    @property
    def name(self):
        return "TensorChangeAboveThreshold"

    @property
    def condition_id(self):
        return WatchpointConditionId.TENSOR_CHANGE_TOO_LARGE.value

    @property
    def param_dict(self):
        param_dict = {
            ParamNameEnum.ABS_MEAN_UPDATE_RATIO_GT.value: self._abs_mean_update_ratio_gt,
            ParamNameEnum.EPSILON.value: self._epsilon
        }
        return param_dict

    @property
    def param_names(self):
        """
        Return the list of parameter names.

        Returns:
            list[str], the parameter names.
        """
        names = [
            ParamNameEnum.ABS_MEAN_UPDATE_RATIO_GT.value,
            ParamNameEnum.EPSILON.value
        ]
        return names


class Watchpoint:
    """
    Watchpoint applies condition to specified tensors.

    .. warning::
        All APIs in this class are experimental prototypes that are subject to
        change or delete.

    Args:
        tensors (Iterable[DebuggerTensor]): The tensors to check.
        condition (ConditionBase): The watch condition to apply to tensors.

    Supported Platforms:
        ``Ascend`` ``GPU``

    Examples:
        >>> from mindinsight.debugger import DumpAnalyzer
        >>> from mindinsight.debugger import TensorTooLargeCondition, Watchpoint
        >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
        >>> tensor_list = my_run.select_tensors(
        ...                                     query_string="Conv",
        ...                                     use_regex=True,
        ...                                     iterations=[0],
        ...                                     ranks=[0],
        ...                                     slots=[0]
        ...                                     )
        >>> watchpoint = Watchpoint(tensors=tensor_list,
        ...                         condition=TensorTooLargeCondition(abs_mean_gt=0.0))
        >>> tensor = list(watchpoint.tensors)[0]
        >>> print(tensor.node.name)
        Default/network-WithLossCell/_backbone-AlexNet/conv1-Conv2d/Cast-op7
        >>> print(watchpoint.condition.name)
        TensorTooLarge
    """

    def __init__(self, tensors, condition):
        validate_tensor_list(tensors, 'tensors')
        validate_type(condition, 'condition', ConditionBase, 'ConditionBase')
        self._tensors = tensors
        self._condition = condition

    @property
    def tensors(self):
        """
        Get tensors to check.

        Returns:
            Iterable[DebuggerTensor], the tensors to check.
        """
        return self._tensors

    @property
    def condition(self):
        """
        Get the watch condition to apply to tensors.

        Returns:
            ConditionBase, the watch condition to apply to tensors.
        """
        return self._condition


class WatchpointHandle:
    """Watchpoint handle."""

    def __init__(self, watchpoint_id, watchpoint):
        validate_type(watchpoint, 'watchpoint', Watchpoint, 'Watchpoint')
        self.watchpoint_id = watchpoint_id
        self.condition = watchpoint.condition
        self.sorted_tensors = self._organize_tensor(watchpoint.tensors)
        self.tensors = watchpoint.tensors

    @staticmethod
    def _get_value(default_map, key, default_value):
        """Get key in default map."""
        value = default_map.get(key)
        if value is None:
            value = default_value
            default_map[key] = value
        return value

    def _organize_tensor(self, tensors):
        """Sort out the tensor and remove the duplication."""
        sorted_tensor = {}
        for tensor in tensors:
            validate_type(tensor, 'tensors', DebuggerTensor, 'List[DebuggerTensor]')
            node_map = self._get_value(sorted_tensor, tensor.iteration, {})
            slot_map = self._get_value(node_map, tensor.node.unique_id, {
                'node': tensor.node,
                'slot_map': {}
            }).get('slot_map')
            slot_map[tensor.slot] = tensor
        return sorted_tensor

    def get_iterations(self):
        """Get iterations to be check in this watchpoint."""
        return list(self.sorted_tensors.keys())

    def need_check(self, tensor):
        """Check if the tensor need to be checked."""
        slot_map = self.sorted_tensors.get(tensor.iteration,
                                           {}).get(tensor.node.unique_id, {}).get('slot_map')
        if slot_map.get(tensor.slot) is not None:
            return True
        return False

    def get_check_nodes(self, iteration):
        """Get check nodes."""
        if iteration is None:
            return {}
        check_nodes = {}
        for node_info in self.sorted_tensors.get(iteration, {}).values():
            node = node_info.get('node')
            node_name = node.full_name_with_graph
            check_node = self._get_value(check_nodes, node_name, {
                "rank_id": [node.rank],
                "is_output": True,
                "root_graph_id": [node.root_graph_id]
            })
            if node.rank not in check_node.get('rank_id'):
                check_node["rank_id"].append(node.rank)
        return check_nodes

    def add_watchpoint(self, iteration, debugger_engine):
        """
        Add watchpoint for the selected iteration.
        """
        check_nodes = self.get_check_nodes(iteration)
        # check if watchpoint must be added for the current iteration
        if check_nodes:
            params = self._get_param_list(debugger_engine.dbg_services_module.Parameter)
            debugger_engine.dbg_service.add_watchpoint(
                watchpoint_id=self.watchpoint_id,
                watch_condition=self.condition.condition_id,
                check_node_list=check_nodes,
                parameter_list=params
            )

    def _get_param_list(self, parameter_class):
        """Get param list."""
        params = []
        set_params = self.condition.param_dict
        for param_name in self.condition.param_names:
            set_value = set_params.get(param_name)
            if set_value is not None:
                param = parameter_class(name=param_name, disabled=False, value=set_value)
            else:
                param = parameter_class(name=param_name, disabled=True, value=0.0)
            params.append(param)
        return params

    def watchpoint_hit_on_no_value(self, iteration):
        """
        Returns list of WatchpointHit if tensors' npy files are missing,
        when error_on_no_value =True
        """
        no_value_hit_list = []
        node_map = self.sorted_tensors.get(iteration)
        if not node_map:
            return no_value_hit_list
        for node_info in node_map.values():
            for tensor in node_info.get('slot_map', {}).values():
                if tensor.has_value() is False:
                    hit_params = []
                    hit_detail = HitDetail(hit_params, self.condition)
                    # 32 means there is no value found
                    error_no_value_code = 32
                    no_value_hit = WatchpointHitImpl(tensor=tensor,
                                                     condition=self.condition,
                                                     hit_detail=hit_detail,
                                                     error_code=error_no_value_code)
                    no_value_hit_list.append(no_value_hit)
        return no_value_hit_list


class WatchpointConditionId(Enum):
    """Watchpoint condition ID."""
    OPERATOR_OVERFLOW = 2
    TENSOR_OVERFLOW = 13
    INITIAL_WEIGHT = 14
    TENSOR_TOO_LARGE = 15
    TENSOR_TOO_SMALL = 16
    TENSOR_ALL_ZERO = 17
    TENSOR_CHANGE_TOO_LARGE = 18
    TENSOR_CHANGE_TOO_SMALL = 19
    UNCHANGED_TENSOR = 20
    TENSOR_RANGE = 21


def validate_tensor_list(param, param_name):
    """Validate list."""
    if not isinstance(param, list):
        raise DebuggerParamTypeError(f"The type of {param_name} should be list of DebuggerTensor. "
                                     f"But the actual type is {type(param)}")
    for i, value in enumerate(param):
        if not isinstance(value, DebuggerTensor):
            raise DebuggerParamTypeError(f"The type of {param_name} should be list of DebuggerTensor. "
                                         f"But the {i} value is {type(value)}.")



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Enums for explain data."""
import enum

class BaseEnum(enum.Enum):

    @classmethod
    def list_members(cls):
        """List all members."""
        return [member.value for member in cls]


class DataManagerStatus(BaseEnum):
    """Data manager status."""
    INIT = 'INIT'
    LOADING = 'LOADING'
    DONE = 'DONE'
    INVALID = 'INVALID'


class ExplainFieldsEnum(BaseEnum):
    """Plugin Name Enum."""
    EXPLAIN = 'explain'
    SAMPLE_ID = 'sample_id'
    BENCHMARK = 'benchmark'
    METADATA = 'metadata'
    GROUND_TRUTH_LABEL = 'ground_truth_label'
    INFERENCE = 'inference'
    EXPLANATION = 'explanation'
    HIERARCHICAL_OCCLUSION = 'hierarchical_occlusion'
    STATUS = 'status'


@enum.unique
class CacheStatus(enum.Enum):
    """Train job cache status."""
    NOT_IN_CACHE = "NOT_IN_CACHE"
    CACHING = "CACHING"
    CACHED = "CACHED"


class ExplanationKeys(enum.Enum):
    """Query type enums."""
    HOC = "hoc_layers"  # HOC: Hierarchical Occlusion, an explanation method we propose
    SALIENCY = "saliency_maps"


class ImageQueryTypes(enum.Enum):
    """Image query type enums."""
    ORIGINAL = 'original'  # Query for the original image
    OUTCOME = 'outcome'  # Query for outcome of HOC explanation
    OVERLAY = 'overlay'  # Query for saliency maps overlay



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Source Handler."""
from collections import defaultdict

from mindinsight.datavisual.common.log import logger as log
from mindinsight.debugger.common.exceptions.exceptions import DebuggerParamTypeError, DebuggerParamValueError
from mindinsight.debugger.stream_handler.base_handler import StreamHandlerBase
from mindinsight.domain.graph.query import StackQuery


class SourceHandler(StreamHandlerBase):
    """Source Data object."""

    def __init__(self):
        # contains the stack info with different file path and line number
        self._stack_info_set = set()
        # keep the stack info in order
        self._stack_infos = []

    def put(self, value):
        """
        Put source object into cache.

        Args:
            value (list[DebuggerSource]): The list of source object.
        """
        if not isinstance(value, list):
            value = [value]
        for source in value:
            # avoid showing annotation which has no file_path
            if source.file_path:
                self._stack_info_set.add(source)

    def sort(self):
        """Sort the stack info according to file path and line number."""
        self._stack_infos = list(self._stack_info_set)
        self._stack_infos.sort()

    def get(self, filter_condition):
        """
        Get stack infos according to filter_condition.

        Args:
            filter_condition (str): The pattern of stack info.

        Returns:
            list, the list of DebuggerSource objects.
        """
        if filter_condition:
            query = StackQuery(self._stack_infos)
            stack_infos = query.filter(filter_condition).all()
        else:
            stack_infos = self._stack_infos
        return stack_infos

    def get_stack_info_by_offset(self, pattern=None, limit=0, offset=0):
        """
        Get stack infos.

        Args:
            pattern (str): The pattern of stack infos. Default: None. If not given, return all stack infos.
            limit (int): The size of each page. Default: 0. If 0, there is no limitation.
            offset (int): The index of the page. Valid only when `limit` is not 0.

        Returns:
            dict, stack info objects. The format is like:
                {
                    'total': int,
                    'offset': int,
                    'stack_infos': [{<file_path>: [{'file_path': str, 'line_no': int, 'code_line': str}]]
                }

        """
        # validate params
        self.check_int('limit', limit, min_num=0, max_num=100)
        self.check_int('offset', offset, min_num=0, max_num=len(self._stack_infos))
        validate_stack_pattern(pattern)
        if not limit and offset > 0:
            return {}
        # get filter results
        filter_res = self.get(pattern)
        if not filter_res:
            log.debug("No stack info with pattern %s", pattern)
            return {}
        merged_res = self._merge_stack_by_file_path(filter_res)
        total_size = len(merged_res)
        if not limit:
            limit = total_size
        st_index = offset * limit
        query_res = merged_res[st_index: st_index + limit]
        for stack_info in query_res:
            source_items = stack_info['items']
            stack_info['items'] = list(map(lambda x: x.to_dict(), source_items))

        return {'total': total_size,
                'offset': offset,
                'stack_infos': query_res}

    @staticmethod
    def check_int(param_name, value, min_num, max_num):
        """Check if the value is positive integer."""
        if isinstance(value, int) and min_num <= value <= max_num:
            return
        log.error("Invalid param `%s`. The integer should be in [%d, %d].", param_name, min_num, max_num)
        raise DebuggerParamTypeError(f"Invalid param `{param_name}`.")

    @staticmethod
    def _merge_stack_by_file_path(stack_infos):
        """
        Merge stack infos by file path.

        Args:
            stack_infos (list[DebuggerSource]): List of Debugger Source objects which is sorted already.

        Returns:
            list, list of merged stack infos. The format of stack info is like:
                {'file_path': str, 'items': [DebuggerSource]}
        """
        merged_dict = defaultdict(list)
        for stack_info in stack_infos:
            merged_dict[stack_info.file_path].append(stack_info)
        res = [None] * len(merged_dict)
        for index, (file_path, source_items) in enumerate(merged_dict.items()):
            res[index] = {'file_path': file_path, 'items': source_items}
        return res


def validate_stack_pattern(stack_pattern):
    """Check stack pattern."""
    if stack_pattern:
        if not isinstance(stack_pattern, str):
            log.error("Invalid stack pattern. String type is required, but got %s.", type(stack_pattern))
            raise DebuggerParamTypeError("stack_pattern is not string type.")
        pattern_limit = 255
        if len(stack_pattern) > pattern_limit:
            log.error("The length of stack_pattern is %s, which should no greater than %s.", len(stack_pattern),
                      pattern_limit)
            raise DebuggerParamValueError("stack_pattern is over length limit.")



# Copyright 2019-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
Management of all events data.

This module exists to all loaders.
It can read events data through the DataLoader.

This module also acts as a thread pool manager.
"""
import abc
import datetime
import threading
import time
import os
from typing import Iterable, Optional

from mindinsight.datavisual.data_transform.summary_watcher import SummaryWatcher

from mindinsight.conf import settings
from mindinsight.datavisual.common import exceptions
from mindinsight.datavisual.common.enums import CacheStatus
from mindinsight.datavisual.common.log import logger
from mindinsight.datavisual.common.enums import DataManagerStatus, BriefCacheStatus
from mindinsight.datavisual.common.enums import PluginNameEnum
from mindinsight.datavisual.common.exceptions import TrainJobNotExistError
from mindinsight.datavisual.data_transform.loader_generators.loader_generator import MAX_DATA_LOADER_SIZE
from mindinsight.datavisual.data_transform.loader_generators.data_loader_generator import DataLoaderGenerator
from mindinsight.utils.computing_resource_mgr import ComputingResourceManager
from mindinsight.utils.exceptions import MindInsightException
from mindinsight.utils.exceptions import ParamValueError
from mindinsight.utils.exceptions import UnknownError
from mindinsight.datavisual.utils.tools import exception_wrapper


class _BasicTrainJob:
    """
    Basic info about train job.

    Args:
        abs_summary_base_dir (str): The canonical path of summary base directory. It should be the return value of
            realpath().
        entry (dict): The summary dir entry listed by SummaryWatcher.
    """
    def __init__(self, abs_summary_base_dir, entry):
        self._abs_summary_base_dir = abs_summary_base_dir
        self._entry = entry

    @property
    def abs_summary_dir(self):
        """Get summary directory path."""
        return os.path.realpath(os.path.join(self._abs_summary_base_dir, self._entry['relative_path']))

    @property
    def summary_base_dir(self):
        """Get summary base directory path."""
        return self._abs_summary_base_dir

    @property
    def train_id(self):
        """Get train id."""
        return self._entry['relative_path']

    @property
    def profiler_dir(self):
        """Get profiler directory path."""
        if self._entry['profiler'] is not None:
            return self._entry['profiler']['directory']
        return None

    @property
    def create_time(self):
        """Get create time."""
        return self._entry['create_time']

    @property
    def update_time(self):
        """Get update time."""
        return self._entry['update_time']

    @property
    def profiler_type(self):
        """Get profiler type"""
        if self._entry['profiler'] is not None:
            return self._entry['profiler']['profiler_type']
        return ''

    @property
    def summary_files(self):
        """Get the summary files count in the summary dir."""
        return self._entry['summary_files']

    @property
    def graph_files(self):
        """Get the graph pb files count in the summary dir."""
        return self._entry['graph_files']

    @property
    def lineage_files(self):
        """Get the lineage files count in the summary dir."""
        return self._entry['lineage_files']

    @property
    def dump_dir(self):
        """Get the dump file path in the summary dir."""
        return self._entry.get('dump_dir', None)


class CachedTrainJob:
    """
    Cache item for BriefCacheManager.

    DetailCacheManager will also wrap it's return value with this class.

    Args:
        basic_info (_BasicTrainJob): Basic info about the train job.
    """
    def __init__(self, basic_info: _BasicTrainJob):
        self._basic_info = basic_info
        self._last_access_time = datetime.datetime.utcnow()

        # Other cached content is stored here.
        self._content = {}

        self._cache_status = CacheStatus.NOT_IN_CACHE
        self._key_locks = {}
        self._is_integrity = True

    @property
    def is_integrity(self):
        return self._is_integrity

    @is_integrity.setter
    def is_integrity(self, integrity):
        self._is_integrity = integrity

    @property
    def cache_status(self):
        """Get cache status."""
        return self._cache_status

    @cache_status.setter
    def cache_status(self, value):
        """Set cache status."""
        self._cache_status = value

    def update_access_time(self):
        """Update last access time of this cache item."""
        self._last_access_time = datetime.datetime.utcnow()

    @property
    def last_access_time(self):
        """Get last access time for purposes such as LRU."""
        return self._last_access_time

    @property
    def abs_summary_dir(self):
        """Get summary directory path."""
        return self._basic_info.abs_summary_dir

    @property
    def summary_base_dir(self):
        """Get summary base directory path."""
        return self._basic_info.summary_base_dir

    def set(self, key, value):
        """Set value to cache."""
        self._content[key] = value

    def delete(self, key, raise_exception=True):
        """Delete key in cache."""
        try:
            self._content.pop(key)
        except KeyError:
            if raise_exception:
                raise ParamValueError("Delete failed. Invalid cache key({}).".format(key))

    def get(self, key, raise_exception=True):
        """
        Get value from cache.

        Args:
            key (str): Key of content.
            raise_exception (bool): If the key does not exist and
                raise_exception is True, it will raise an Exception.

        Returns:
            Union[Object, None], Return value if key in content,
                return False else if raise_exception is False.
        Raises:
            ParamValueError, if the key does not exist and raise_exception is True.

        """
        try:
            return self._content[key]
        except KeyError:
            if raise_exception:
                raise ParamValueError("Invalid cache key({}).".format(key))
            return None

    @property
    def basic_info(self):
        """Get basic train job info."""
        return self._basic_info

    @basic_info.setter
    def basic_info(self, value):
        """Set basic train job info."""
        self._basic_info = value

    def lock_key(self, key):
        """Threading lock with given key."""
        return self._key_locks.setdefault(key, threading.Lock())

    @property
    def train_id(self):
        """Get train id."""
        return self._basic_info.train_id


class TrainJob:
    """
    Train job object.

    You must not create TrainJob objects manually. You should always get TrainJob objects from DataManager.

    Args:
        brief_train_job (CachedTrainJob): Brief info about train job.
        detail_train_job (Optional[CachedTrainJob]): Detailed info about train job. Default: None.
    """
    def __init__(self,
                 brief_train_job: CachedTrainJob,
                 detail_train_job: Optional[CachedTrainJob] = None):
        self._brief = brief_train_job
        self._detail = detail_train_job
        if self._detail is None:
            self._cache_status = CacheStatus.NOT_IN_CACHE
            self._is_integrity = True
        else:
            self._cache_status = self._detail.cache_status
            self._is_integrity = self._detail.is_integrity

    @property
    def is_integrity(self):
        return self._is_integrity

    @is_integrity.setter
    def is_integrity(self, integrity):
        self._is_integrity = integrity

    def has_detail(self):
        """Whether this train job has detailed info in cache."""
        return bool(self._detail is not None)

    def get_detail(self, key):
        """
        Get detail content.

        Args:
            key (Any): Cache key.

        Returns:
            Any, cache content.

        Raises:
            TrainJobDetailNotInCacheError: when this train job has no detail cache.

        """
        if not self.has_detail():
            raise exceptions.TrainJobDetailNotInCacheError()
        return self._detail.get(key)

    def get_brief(self, key):
        """
        Get brief content.

        Args:
            key (Any): Cache key.

        Returns:
            Any, cache content.
        """
        return self._brief.get(key)

    def get_basic_info(self):
        """
        Get basic info.

        Returns:
            basic_info (_BasicTrainJob): Basic info about the train job.
        """
        return self._brief.basic_info

    @property
    def cache_status(self):
        """Get cache status."""
        return self._cache_status

    @cache_status.setter
    def cache_status(self, cache_status):
        """Set cache status."""
        self._cache_status = cache_status


class BaseCacheItemUpdater(abc.ABC):
    """Abstract base class for other modules to update cache content."""
    def update_item(self, cache_item: CachedTrainJob):
        """
        Update cache item in place.

        Args:
            cache_item (CachedTrainJob): The cache item to be processed.
        """
        raise NotImplementedError()


class _BaseCacheManager:
    """Base class for cache manager."""

    def __init__(self, summary_base_dir):
        self._summary_base_dir = summary_base_dir
        # Use dict to remove duplicate updaters.
        self._updaters = {}

        # key is train_id
        self._lock = threading.Lock()
        self._cache_items = {}

    def size(self):
        """Gets used cache slots."""
        return len(self._cache_items)

    def register_cache_item_updater(self, updater: BaseCacheItemUpdater):
        """Register cache item updater."""
        self._updaters[updater.__class__.__qualname__] = updater

    def get_train_jobs(self):
        """Get cached train jobs."""
        copied_train_jobs = dict(self._cache_items)
        return copied_train_jobs

    def get_train_job(self, train_id):
        """Get cached train job."""
        try:
            return self._cache_items[train_id]
        except KeyError:
            raise TrainJobNotExistError(train_id)

    def cache_train_job(self, train_id) -> bool:
        """
        Cache given train job and update train job's last access time.

        This method should return true if reload actions should be taken to cache the train job.

        Args:
            train_id (str): Train Id.
        """
        raise NotImplementedError()

    def delete_train_job(self, train_id):
        """Delete train job from cache."""
        if train_id in self._cache_items:
            del self._cache_items[train_id]

    def has_content(self):
        """Whether this cache manager has train jobs."""
        return bool(self._cache_items)

    def update_cache(self, executor):
        """
        Update cache according to given train jobs on disk.

        Different cache manager should implement different cache update policies in this method.

        Args:
            executor (Executor): The Executor instance.
        """
        raise NotImplementedError()


class _BriefCacheManager(_BaseCacheManager):
    """A cache manager that holds all disk train jobs on disk."""

    def __init__(self, summary_base_dir):
        super(_BriefCacheManager, self).__init__(summary_base_dir)
        self._summary_watcher = SummaryWatcher()
        self._status = BriefCacheStatus.INIT.value

    @property
    def status(self):
        """
        Get the status of brief cache manager.

        Returns:
            BriefCacheStatus, the status of brief cache manager.
        """
        return self._status

    @status.setter
    def status(self, status):
        """Set brief cache manager status."""
        self._status = status

    def cache_train_job(self, train_id):
        """
        Cache given train job.

        All disk train jobs are cached on every reload, so this method always return false.

        Args:
            train_id (str): Train Id.
        """
        if train_id in self._cache_items:
            self._cache_items[train_id].update_access_time()

        return False

    def update_cache(self, executor=None):
        """Update cache."""
        logger.debug('Start to update BriefCacheManager.')
        summaries_info = self._summary_watcher.list_summary_directories(self._summary_base_dir)

        basic_train_jobs = []
        for info in summaries_info:
            basic_train_jobs.append(_BasicTrainJob(
                abs_summary_base_dir=self._summary_base_dir,
                entry=info
            ))

        with self._lock:
            new_cache_items = self._merge_with_disk(basic_train_jobs)
            self._cache_items = new_cache_items
        for updater in self._updaters.values():
            for cache_item in self._cache_items.values():
                updater.update_item(cache_item)

    def _merge_with_disk(self, disk_train_jobs: Iterable[_BasicTrainJob]):
        """
        Merge train jobs in cache with train jobs from disk

        This method will remove train jobs not on disk. Call this function with lock for thread safety.

        Args:
            disk_train_jobs (Iterable[_BasicTrainJob]): Basic train jobs info from disk.

        Returns:
            dict, a dict containing train jobs to be cached.
        """
        new_cache_items = {}
        for train_job in disk_train_jobs:
            if train_job.train_id not in self._cache_items:
                new_cache_items[train_job.train_id] = CachedTrainJob(train_job)
            else:
                reused_train_job = self._cache_items[train_job.train_id]
                reused_train_job.basic_info = train_job
                new_cache_items[train_job.train_id] = reused_train_job

        return new_cache_items

    def register_folder_analyzer(self, analyzer):
        """Register folder analyzer."""
        self._summary_watcher.register_folder_analyzer(analyzer)

    @property
    def cache_items(self):
        """Get cache items."""
        return self._cache_items


# Key for plugin tags.
DATAVISUAL_PLUGIN_KEY = "tag_mapping"
# Detail train job cache key for datavisual content.
DATAVISUAL_CACHE_KEY = "datavisual"


class _DetailCacheManager(_BaseCacheManager):
    """A cache manager that holds detailed info for most recently used train jobs."""
    def __init__(self, summary_base_dir):
        super().__init__(summary_base_dir)
        self._loader_pool = {}
        self._deleted_id_list = []
        self._loader_pool_mutex = threading.Lock()
        self._loader_generators = [DataLoaderGenerator(summary_base_dir)]
        self._loading_mutex = threading.Lock()
        self.summaries_info = []

    def has_content(self):
        """Whether this cache manager has train jobs."""
        return bool(self._loader_pool)

    def register_folder_analyzer(self, analyzer):
        """Register folder analyzer."""
        for generator in self._loader_generators:
            generator.register_folder_analyzer(analyzer)

    def size(self):
        """
        Get the number of items in this cache manager.

        To be implemented.

        Returns:
            int, the number of items in this cache manager.
        """
        raise NotImplementedError()

    def loader_pool_size(self):
        """Get loader pool size."""
        return len(self._loader_pool)

    def update_cache(self, executor):
        """
        Update cache.

        Will switch to using disk_train_jobs in the future.

        Args:
            executor (Executor): The Executor instance.
        """
        with self._loading_mutex:
            load_in_cache = exception_wrapper(self._execute_load_data)
            try:
                while not load_in_cache(executor):
                    yield
            except UnknownError as ex:
                logger.warning("Load event data failed. Detail: %s.", str(ex))

    def cache_train_job(self, train_id):
        """Cache given train job."""
        loader = None
        need_reload = False
        with self._loader_pool_mutex:
            if self._is_loader_in_loader_pool(train_id, self._loader_pool):
                loader = self._loader_pool.get(train_id)

            if loader is None:
                for generator in self._loader_generators:
                    tmp_loader = generator.generate_loader_by_train_id(train_id)
                    if loader and loader.latest_update_time > tmp_loader.latest_update_time:
                        continue
                    loader = tmp_loader

                if loader is None:
                    raise TrainJobNotExistError(train_id)

                self._add_loader(loader)
                need_reload = True

        self._update_loader_latest_update_time(loader.loader_id)
        return need_reload

    def get_train_jobs(self):
        """
        Get train jobs

        To be implemented.
        """

    def _add_loader(self, loader):
        """
        Add a loader to load data.

        Args:
            loader (LoaderStruct): A object of `Loader`.
        """
        if len(self._loader_pool) >= MAX_DATA_LOADER_SIZE:
            delete_number = len(self._loader_pool) - MAX_DATA_LOADER_SIZE + 1
            sorted_loaders = sorted(self._loader_pool.items(),
                                    key=lambda loader: loader[1].latest_update_time)
            for index in range(delete_number):
                delete_loader_id = sorted_loaders[index][0]
                self._delete_loader(delete_loader_id)
        self._loader_pool.update({loader.loader_id: loader})

    def _delete_loader(self, loader_id):
        """
        Delete loader from loader pool by loader id.

        Args:
            loader_id (str): ID of loader.
        """
        if self._loader_pool.get(loader_id) is not None:
            logger.debug("delete loader %s", loader_id)
            self._loader_pool.pop(loader_id)

    def _execute_loader(self, loader_id, executor):
        """
        Load data form data_loader.

        If there is something wrong by loading, add logs and delete the loader.

        Args:
            loader_id (str): An ID for `Loader`.
            executor (Executor): The Executor instance.

        Returns:
            bool, True if the loader is finished loading.
        """
        try:
            with self._loader_pool_mutex:
                loader = self._loader_pool.get(loader_id, None)
                if loader is None:
                    logger.debug("Loader %r has been deleted, will not load data.", loader_id)
                    return True

            loader.cache_status = CacheStatus.CACHING
            if loader.data_loader.load(executor):
                # Update loader cache status to CACHED.
                # Loader with cache status CACHED should remain the same cache status.
                loader.cache_status = CacheStatus.CACHED
                return True
            return False

        except MindInsightException as ex:
            logger.warning("Data loader %r load data failed. "
                           "Delete data_loader. Detail: %s", loader_id, ex)

            with self._loader_pool_mutex:
                self._delete_loader(loader_id)
            return True

    def _generate_loaders(self):
        """This function generates the loader from given path."""
        loader_dict = {}
        for generator in self._loader_generators:
            loader_dict.update(generator.generate_loaders(self._loader_pool, self.summaries_info))

        sorted_loaders = sorted(loader_dict.items(), key=lambda loader: loader[1].latest_update_time)
        latest_loaders = sorted_loaders[-MAX_DATA_LOADER_SIZE:]
        self._deal_loaders(latest_loaders)

    def _deal_loaders(self, latest_loaders):
        """
        This function determines which loaders to keep or remove or added.

        It is based on the given dict of loaders.

        Args:
            latest_loaders (list[dict]): A list of <loader_id: LoaderStruct>.
        """

        with self._loader_pool_mutex:
            for loader_id, loader in latest_loaders:
                if self._loader_pool.get(loader_id, None) is None:
                    self._add_loader(loader)
                    continue

                # If this loader was updated manually before,
                # its latest_update_time may bigger than update_time in summary.
                if self._loader_pool[loader_id].latest_update_time < loader.latest_update_time:
                    self._update_loader_latest_update_time(loader_id, loader.latest_update_time)

    def _execute_load_data(self, executor):
        """Load data through multiple threads."""
        self._generate_loaders()
        loader_pool = self._get_snapshot_loader_pool()
        loaded = True
        for loader_id in loader_pool:
            loaded = self._execute_loader(loader_id, executor) and loaded
        return loaded

    def delete_train_job(self, train_id):
        """
        Delete train job with a train id.

        Args:
            train_id (str): ID for train job.

        """
        with self._loader_pool_mutex:
            self._delete_loader(train_id)

    def list_tensors(self, train_id, tag):
        """
        List tensors of the given train job and tag.

        If the tensor can not find by the given tag, will raise exception.

        Args:
            train_id (str): ID for train job.
            tag (str): The tag name.

        Returns:
            list, the NameTuple format is `collections.namedtuple('_Tensor', ['wall_time', 'event_step', 'value'])`.
                the value will contain the given tag data.

        """
        loader_pool = self._get_snapshot_loader_pool()
        if not self._is_loader_in_loader_pool(train_id, loader_pool):
            raise TrainJobNotExistError("Can not find the given train job in cache.")

        data_loader = loader_pool[train_id].data_loader

        tensors = []
        try:
            events_data = data_loader.get_events_data()
            tensors = events_data.tensors(tag)
        except KeyError:
            error_msg = "Can not find any data in this train job by given tag."
            raise ParamValueError(error_msg)
        except AttributeError:
            logger.debug("Train job %r has been deleted or it has not loaded data, "
                         "and set tags to empty list.", train_id)

        return tensors

    def _check_train_job_exist(self, train_id, loader_pool):
        """
        Check train job exist, if not exist, will raise exception.

        Args:
            train_id (str): The given train job id.
            loader_pool (dict[str, LoaderStruct]): Refer to self._loader_pool.

        Raises:
            TrainJobNotExistError: Can not find train job in data manager.
        """
        is_exist = False
        if train_id in loader_pool:
            return
        for generator in self._loader_generators:
            if generator.check_train_job_exist(train_id):
                is_exist = True
                break
        if not is_exist:
            raise TrainJobNotExistError("Can not find the train job in data manager.")

    def _is_loader_in_loader_pool(self, train_id, loader_pool):
        """
        Check train job exist, if not exist, return False. Else, return True.

        Args:
            train_id (str): The given train job id.
            loader_pool (dict): See self._loader_pool.

        Returns:
            bool, if loader in loader pool, return True.
        """
        if train_id in loader_pool:
            return True
        return False

    def _get_snapshot_loader_pool(self):
        """
        Create a snapshot of data loader pool to avoid concurrent mutation and iteration issues.

        Returns:
            dict, a copy of `self._loader_pool`.
        """
        with self._loader_pool_mutex:
            return dict(self._loader_pool)

    def get_train_job(self, train_id):
        """
        Get train job by train ID.

        This method overrides parent method.

        Args:
            train_id (str): Train ID for train job.
        Returns:
            dict, single train job, if can not find any data, will return None.
        """
        self._check_train_job_exist(train_id, self._loader_pool)

        loader = self._get_loader(train_id)
        if loader is None:
            logger.info("No valid summary log in train job %s, or it is not in the cache.", train_id)
            return None

        train_job = loader.to_dict()
        train_job.pop('data_loader')

        plugin_data = {}
        for plugin_name in PluginNameEnum.list_members():
            job = self.get_train_job_by_plugin(train_id, plugin_name=plugin_name)
            if job is None:
                plugin_data[plugin_name] = []
            else:
                plugin_data[plugin_name] = job['tags']

        train_job.update({DATAVISUAL_PLUGIN_KEY: plugin_data})

        # Will fill basic_info value in future.
        train_job_obj = CachedTrainJob(basic_info=None)
        train_job_obj.set(DATAVISUAL_CACHE_KEY, train_job)

        train_job_obj.cache_status = loader.cache_status
        train_job_obj.is_integrity = loader.data_loader.is_integrity
        return train_job_obj

    def _get_loader(self, train_id):
        """
        Get loader by train id.

        Args:
            train_id (str): Train Id.

        Returns:
            LoaderStruct, the loader.
        """
        loader = None
        with self._loader_pool_mutex:
            if self._is_loader_in_loader_pool(train_id, self._loader_pool):
                loader = self._loader_pool.get(train_id)

        return loader

    def _update_loader_latest_update_time(self, loader_id, latest_update_time=None):
        """
        Update loader with latest_update_time.

        Args:
            loader_id (str): ID of loader.
            latest_update_time (float): Timestamp.
        """
        if latest_update_time is None:
            latest_update_time = time.time()
        self._loader_pool[loader_id].latest_update_time = latest_update_time

    def get_train_job_by_plugin(self, train_id, plugin_name):
        """
        Get a train job by train job id.

        If the given train job does not has the given plugin data, the tag list will be empty.

        Args:
            train_id (str): Get train job info by the given id.
            plugin_name (str): Get tags by given plugin.

        Returns:
            TypedDict('TrainJobEntity', {'id': str, 'name': str, 'tags': List[str]}),
                a train job object.

        """
        self._check_train_job_exist(train_id, self._loader_pool)

        loader = self._get_loader(train_id)
        if loader is None:
            logger.warning("No valid summary log in train job %s, "
                           "or it is not in the cache.", train_id)
            return None

        name = loader.name
        data_loader = loader.data_loader

        tags = []
        try:
            events_data = data_loader.get_events_data()
            tags = events_data.list_tags_by_plugin(plugin_name)
        except KeyError:
            logger.debug("Plugin name %r does not exist "
                         "in train job %r, and set tags to empty list.", plugin_name, name)
        except AttributeError:
            logger.debug("Train job %r has been deleted or it has not loaded data, "
                         "and set tags to empty list.", name)

        result = dict(id=train_id, name=name, tags=tags)
        return result


class DataManager:
    """
    DataManager manages a pool of loader which help access events data.

    Each loader helps deal the data of the events.
    A loader corresponds to an events_data.
    The DataManager build a pool including all the data_loader.
    The data_loader provides extracting
    method to get the information of events.
    """
    def __init__(self, summary_base_dir):
        """
        Initialize the pool of loader and the dict of name-to-path.

        Args:
            summary_base_dir (str): Base summary directory.

        self._status: Refer `datavisual.common.enums.DataManagerStatus`.

        """
        self._summary_base_dir = os.path.realpath(summary_base_dir)
        self._status = DataManagerStatus.INIT.value
        self._status_mutex = threading.Lock()
        self._brief_cache_status_mutex = threading.Lock()

        self._detail_cache = _DetailCacheManager(self._summary_base_dir)
        self._brief_cache = _BriefCacheManager(self._summary_base_dir)

        # This lock is used to make sure that only one self._load_data_in_thread() is running.
        # Because self._load_data_in_thread() will create process pool when loading files, we can not
        # afford to run multiple self._load_data_in_thread() simultaneously (will create too many processes).
        self._load_data_lock = threading.Lock()
        self._load_brief_data_lock = threading.Lock()
        self._summary_watcher = SummaryWatcher()

    @property
    def summary_base_dir(self):
        """Get summary base dir."""
        return self._summary_base_dir

    def start_load_data(self, reload_interval=0):
        """
        Start threads for loading data.

        Args:
            reload_interval (int): Time to reload data again.

        Returns:
            Thread, the background Thread instance.
        """
        logger.info("Start to load data")
        DataManager.check_reload_interval(reload_interval)
        thread = threading.Thread(target=self._load_data_in_thread,
                                  name='start_load_data_thread',
                                  args=(reload_interval,),
                                  daemon=True)
        thread.start()

        load_brief_data_thread = threading.Thread(target=self._load_brief_data_loop,
                                                  name='start_load_brief_data_thread',
                                                  args=(reload_interval,),
                                                  daemon=True)
        load_brief_data_thread.start()
        return thread, load_brief_data_thread

    @staticmethod
    def check_reload_interval(reload_interval):
        """
        Check reload interval is valid.

        Args:
            reload_interval (int): Reload interval >= 0.
        """
        if not isinstance(reload_interval, int):
            raise ParamValueError("The value of reload interval should be integer.")

        if reload_interval < 0:
            raise ParamValueError("The value of reload interval should be >= 0.")

    def _load_brief_data_loop(self, reload_interval):
        """Wrapper for load brief data in thread."""
        if self._load_brief_data_lock.locked():
            return
        with self._load_brief_data_lock:
            while True:
                try:
                    exception_wrapper(self._load_brief_data)()
                except UnknownError as exc:
                    # Not raising the exception here to ensure that data reloading does not crash.
                    logger.warning(exc.message)
                finally:
                    self._brief_cache.status = BriefCacheStatus.LOADED.value
                if not reload_interval:
                    break
                time.sleep(reload_interval)

    def _load_data_in_thread(self, reload_interval):
        """Wrapper for load data in thread."""
        if self._load_data_lock.locked():
            return
        with self._load_data_lock:
            while True:
                try:
                    exception_wrapper(self._load_data)()
                except UnknownError as exc:
                    # Not raising the exception here to ensure that data reloading does not crash.
                    logger.warning(exc.message)
                finally:
                    self._status = DataManagerStatus.DONE.value
                if not reload_interval:
                    break
                time.sleep(reload_interval)

    def _load_brief_data(self):
        """This function will load data once and ignore it if the status is loading."""
        with self._brief_cache_status_mutex:
            if self._brief_cache.status == BriefCacheStatus.LOADING.value:
                logger.debug("Current brief cache status is %s , will ignore to load data.", self._brief_cache.status)
                return
            self._brief_cache.status = BriefCacheStatus.LOADING.value

        self._brief_cache.update_cache()

        with self._brief_cache_status_mutex:
            self._brief_cache.status = BriefCacheStatus.LOADED.value
            logger.debug("Load brief data end.")

    def _load_data(self):
        """This function will load data once and ignore it if the status is loading."""
        with self._status_mutex:
            if self.status == DataManagerStatus.LOADING.value:
                logger.debug("Current status is %s , will ignore to load data.", self.status)
                return
            self.status = DataManagerStatus.LOADING.value

        with ComputingResourceManager.get_instance().get_executor(
                max_processes_cnt=settings.MAX_PROCESSES_COUNT) as executor:
            self._detail_cache.summaries_info = self._summary_watcher.list_summary_directories(self._summary_base_dir)
            last_time = time.time()
            for _ in self._detail_cache.update_cache(executor):
                update_interval = time.time() - last_time
                logger.debug('Loading one round of detail cache taking %ss.', update_interval)
                last_time = time.time()
            with self._status_mutex:
                if not self._brief_cache.has_content() and not self._detail_cache.has_content():
                    self.status = DataManagerStatus.INVALID.value
                else:
                    self.status = DataManagerStatus.DONE.value

                logger.info("Load detail data end, and loader pool size is %r.", self._detail_cache.loader_pool_size())

    def get_train_job_by_plugin(self, train_id, plugin_name):
        """
        Get a train job by train job id.

        If the given train job does not has the given plugin data, the tag list will be empty.

        Args:
            train_id (str): Get train job info by the given id.
            plugin_name (str): Get tags by given plugin.

        Returns:
            TypedDict('TrainJobEntity', {'id': str, 'name': str, 'tags': List[str]}),
                a train job object.

        """
        self._check_status_valid()
        return self._detail_cache.get_train_job_by_plugin(train_id, plugin_name)

    def delete_train_job(self, train_id, only_delete_from_cache=True):
        """
        Delete train job with a train id.

        Args:
            train_id (str): ID for train job.

        """
        if not only_delete_from_cache:
            raise NotImplementedError("Delete from both cache and disk is not supported.")

        self._brief_cache.delete_train_job(train_id)
        self._detail_cache.delete_train_job(train_id)

    def list_tensors(self, train_id, tag):
        """
        List tensors of the given train job and tag.

        If the tensor can not find by the given tag, will raise exception.

        Args:
            train_id (str): ID for train job.
            tag (str): The tag name.

        Returns:
            NamedTuple, the tuple format is `collections.namedtuple('_Tensor', ['wall_time', 'event_step', 'value'])`.
                the value will contain the given tag data.

        """
        self._check_status_valid()
        return self._detail_cache.list_tensors(train_id, tag)

    def _check_status_valid(self):
        """Check if the status is valid to load data."""

        if self.status == DataManagerStatus.INIT.value:
            raise exceptions.SummaryLogIsLoading("Data is being loaded, current status: %s." % self._status)

    def get_train_job(self, train_id):
        """
        Get train job by train ID.

        Args:
            train_id (str): Train ID for train job.

        Returns:
            dict, single train job, if can not find any data, will return None.
        """
        self._check_status_valid()
        detail_train_job = self._detail_cache.get_train_job(train_id)
        brief_train_job = self._brief_cache.get_train_job(train_id)

        return TrainJob(brief_train_job, detail_train_job)

    @property
    def status(self):
        """
        Get the status of data manager.

        Returns:
            DataManagerStatus, the status of data manager.
        """
        return self._status

    @status.setter
    def status(self, status):
        """Set data manager status."""
        self._status = status

    def cache_train_job(self, train_id):
        """Cache given train job (async)."""
        brief_need_reload = self._brief_cache.cache_train_job(train_id)
        detail_need_reload = self._detail_cache.cache_train_job(train_id)
        if brief_need_reload or detail_need_reload:
            self.start_load_data()

    def register_brief_cache_item_updater(self, updater: BaseCacheItemUpdater):
        """Register brief cache item updater for brief cache manager."""
        self._brief_cache.register_cache_item_updater(updater)

    def register_folder_analyzer(self, analyzer):
        """Register folder analyzer."""
        self._brief_cache.register_folder_analyzer(analyzer)
        self._detail_cache.register_folder_analyzer(analyzer)
        self._summary_watcher.register_folder_analyzer(analyzer)

    def get_brief_cache(self):
        """Get brief cache."""
        return self._brief_cache

    def get_brief_train_job(self, train_id):
        """Get brief train job."""
        return self._brief_cache.get_train_job(train_id)


DATA_MANAGER = DataManager(settings.SUMMARY_BASE_DIR)



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
File parser for MindExplain data.

This module is used to parse the MindExplain log file.
"""
from collections import namedtuple

from mindinsight.datavisual.common import exceptions
from mindinsight.datavisual.data_access.file_handler import FileHandler
from mindinsight.datavisual.data_transform.ms_data_loader import _SummaryParser
from mindinsight.datavisual.proto_files import mindinsight_xai_pb2 as xai_pb2
from mindinsight.explainer.common.enums import ExplainFieldsEnum
from mindinsight.explainer.common.log import logger
from mindinsight.utils.exceptions import UnknownError

HEADER_SIZE = 8
CRC_STR_SIZE = 4
MAX_EVENT_STRING = 500000000

BenchmarkContainer = namedtuple('BenchmarkContainer', ['benchmark', 'status'])
MetadataContainer = namedtuple('MetadataContainer', ['metadata', 'status'])
InferfenceContainer = namedtuple('InferenceContainer', ['ground_truth_prob',
                                                        'ground_truth_prob_sd',
                                                        'ground_truth_prob_itl95_low',
                                                        'ground_truth_prob_itl95_hi',
                                                        'predicted_label',
                                                        'predicted_prob',
                                                        'predicted_prob_sd',
                                                        'predicted_prob_itl95_low',
                                                        'predicted_prob_itl95_hi'])
SampleContainer = namedtuple('SampleContainer', ['sample_id', 'image_path', 'ground_truth_label', 'inference',
                                                 'explanation', 'hierarchical_occlusion', 'status'])


class ExplainParser(_SummaryParser):
    """The summary file parser."""

    def __init__(self, summary_dir):
        super(ExplainParser, self).__init__(summary_dir)
        self._latest_offset = 0

    def list_events(self, filenames):
        """
        Load summary file and parse file content.

        Args:
            filenames (list[str]): File name list.

        Returns:
            tuple, the elements of the tuple are:

                - file_changed (bool): True if the latest file is changed.
                - is_end (bool): True if all the summary files are finished loading.
                - event_data (dict): Event data where keys are explanation field.
        """
        summary_files = self.sort_files(filenames)

        is_end = False
        file_changed = False
        event_data = {}
        filename = summary_files[-1]

        file_path = FileHandler.join(self._summary_dir, filename)
        if filename != self._latest_filename:
            self._summary_file_handler = FileHandler(file_path, 'rb')
            self._latest_filename = filename
            self._latest_offset = 0
            file_changed = True

        new_size = FileHandler.file_stat(file_path).size
        if new_size == self._latest_offset:
            is_end = True
            return file_changed, is_end, event_data

        while True:
            start_offset = self._summary_file_handler.offset
            try:
                event_str = self.event_load(self._summary_file_handler)
                if event_str is None:
                    self._summary_file_handler.reset_offset(start_offset)
                    is_end = True
                    return file_changed, is_end, event_data
                if len(event_str) > MAX_EVENT_STRING:
                    logger.warning("file_path: %s, event string: %d exceeds %d and drop it.",
                                   self._summary_file_handler.file_path, len(event_str), MAX_EVENT_STRING)
                    continue

                field_list, tensor_value_list = self._event_decode(event_str)
                for field, tensor_value in zip(field_list, tensor_value_list):
                    event_data[field] = tensor_value

                logger.debug("Parse summary file offset %d, file path: %s.",
                             self._summary_file_handler.offset, file_path)
                return file_changed, is_end, event_data
            except exceptions.CRCLengthFailedError as ex:
                self._summary_file_handler.reset_offset(start_offset)
                is_end = True
                logger.warning("Check crc failed and reset offset, file_path=%s, offset=%s. Detail: %r.",
                               self._summary_file_handler.file_path, self._summary_file_handler.offset, str(ex))
                return file_changed, is_end, event_data
            except Exception as ex:
                # Note: If an unknown error occurs, we will set the offset to the end of this file,
                # which is equivalent to stopping parsing this file. We do not delete the current job
                # and retain the data that has been successfully parsed.
                self._summary_file_handler.reset_offset(new_size)

                # Notice: If the current job is the latest one in the loader pool and the job is deleted,
                # the job goes into an infinite cycle of load-fail-delete-reload-load-fail-delete.
                # We need to prevent this infinite loop.
                logger.error("Parse summary file failed, will set offset to the file end. file_path: %s, "
                             "offset: %d, detail: %s.", file_path, self._summary_file_handler.offset, str(ex))
                logger.exception(ex)
                raise UnknownError(str(ex))
            finally:
                self._latest_offset = self._summary_file_handler.offset

    @staticmethod
    def _event_decode(event_str):
        """
        Transform `Event` data to tensor_event and update it to EventsData.

        Args:
            event_str (str): Message event string in summary proto, data read from file handler.

        Returns:
            tuple, the elements of the result tuple are:

                - field_list (list): Explain fields to be parsed.
                - tensor_value_list (list): Parsed data with respect to the field list.
        """

        logger.debug("Start to parse event string. Event string len: %s.", len(event_str))
        event = xai_pb2.Event.FromString(event_str)
        logger.debug("Deserialize event string completed.")

        fields = {
            'sample_id': ExplainFieldsEnum.SAMPLE_ID,
            'benchmark': ExplainFieldsEnum.BENCHMARK,
            'metadata': ExplainFieldsEnum.METADATA
        }

        tensor_event_value = getattr(event, 'explain')

        field_list = []
        tensor_value_list = []
        for field in fields:
            if getattr(tensor_event_value, field, None) is None:
                continue

            if ExplainFieldsEnum.METADATA.value == field and not tensor_event_value.metadata.label:
                continue

            tensor_value = None
            if field == ExplainFieldsEnum.SAMPLE_ID.value:
                tensor_value = ExplainParser._add_image_data(tensor_event_value)
            elif field == ExplainFieldsEnum.BENCHMARK.value:
                tensor_value = ExplainParser._add_benchmark(tensor_event_value)
            elif field == ExplainFieldsEnum.METADATA.value:
                tensor_value = ExplainParser._add_metadata(tensor_event_value)
            logger.debug("Event generated, label is %s, step is %s.", field, event.step)
            field_list.append(field)
            tensor_value_list.append(tensor_value)
        return field_list, tensor_value_list

    @staticmethod
    def _add_image_data(tensor_event_value):
        """
        Parse image data based on sample_id in Explain message.

        Args:
            tensor_event_value (Event): The object of Explain message.

        Returns:
            SampleContainer, a named tuple containing sample data.
        """
        inference = InferfenceContainer(
            ground_truth_prob=tensor_event_value.inference.ground_truth_prob,
            ground_truth_prob_sd=tensor_event_value.inference.ground_truth_prob_sd,
            ground_truth_prob_itl95_low=tensor_event_value.inference.ground_truth_prob_itl95_low,
            ground_truth_prob_itl95_hi=tensor_event_value.inference.ground_truth_prob_itl95_hi,
            predicted_label=tensor_event_value.inference.predicted_label,
            predicted_prob=tensor_event_value.inference.predicted_prob,
            predicted_prob_sd=tensor_event_value.inference.predicted_prob_sd,
            predicted_prob_itl95_low=tensor_event_value.inference.predicted_prob_itl95_low,
            predicted_prob_itl95_hi=tensor_event_value.inference.predicted_prob_itl95_hi
        )
        sample_data = SampleContainer(
            sample_id=tensor_event_value.sample_id,
            image_path=tensor_event_value.image_path,
            ground_truth_label=tensor_event_value.ground_truth_label,
            inference=inference,
            explanation=tensor_event_value.explanation,
            hierarchical_occlusion=tensor_event_value.hoc,
            status=tensor_event_value.status
        )
        return sample_data

    @staticmethod
    def _add_benchmark(tensor_event_value):
        """
        Parse benchmark data from Explain message.

        Args:
            tensor_event_value (Event): The object of Explain message.

        Returns:
            BenchmarkContainer, a named tuple containing benchmark data.
        """
        benchmark_data = BenchmarkContainer(
            benchmark=tensor_event_value.benchmark,
            status=tensor_event_value.status
        )

        return benchmark_data

    @staticmethod
    def _add_metadata(tensor_event_value):
        """
        Parse  metadata from Explain message.

        Args:
            tensor_event_value (Event): The object of Explain message.

        Returns:
            MetadataContainer, a named tuple containing benchmark data.
        """
        metadata_value = MetadataContainer(
            metadata=tensor_event_value.metadata,
            status=tensor_event_value.status
        )

        return metadata_value



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Utils for calculate importance."""
import numpy as np

from mindinsight.optimizer.common.exceptions import SamplesNotEnoughError, CorrelationNanError
from mindinsight.optimizer.common.log import logger


def calc_hyper_param_importance(df, hyper_param, target):
    """Calc hyper param importance relative to given target."""
    logger.debug("Calculating importance for hyper_param %s, target is %s.", hyper_param, target)

    new_df = df[[hyper_param, target]]
    no_missing_value_df = new_df.dropna()

    # Can not calc pearson correlation coefficient when number of samples is less or equal than 2
    if len(no_missing_value_df) <= 2:
        raise SamplesNotEnoughError("Number of samples is less or equal than 2.")

    correlation = no_missing_value_df[target].corr(no_missing_value_df[hyper_param])
    if np.isnan(correlation):
        logger.warning("Correlation is nan!")
        raise CorrelationNanError("Correlation is nan!")
    return abs(correlation)



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""BaseNetwork module."""
import os

from jinja2 import Template

from mindinsight.wizard.base.source_file import SourceFile


def render_template(template_file_path, context):
    with open(template_file_path, encoding='utf-8') as fp:
        content = fp.read()
    template = Template(content, trim_blocks=True, lstrip_blocks=True, keep_trailing_newline=True)
    return template.render(context)


class TemplateManager:
    """BaseNetwork code generator."""
    replace_template_suffixes = [('.py-tpl', '.py'), ('.sh-tpl', '.sh'), ('.md-tpl', '.md')]

    def __init__(self, template_base_dir, exclude_dirs=None, exclude_files=None):
        self.template_base_dir = template_base_dir
        self.exclude_dirs = self._get_exclude_paths(template_base_dir, exclude_dirs)
        self.exclude_files = self._get_exclude_paths(template_base_dir, exclude_files)

    @staticmethod
    def _get_exclude_paths(base_dir, exclude_paths):
        """Convert exclude path to absolute directory path."""
        exclude_abs_paths = []
        if exclude_paths is None:
            return exclude_abs_paths

        for exclude_path in exclude_paths:
            if exclude_path.startswith(base_dir):
                exclude_abs_paths.append(exclude_path)
            else:
                exclude_abs_paths.append(os.path.join(base_dir, exclude_path))
        return exclude_abs_paths

    def get_template_files(self):
        """Get template files for template directory."""
        template_files = []
        for root, sub_dirs, files in os.walk(self.template_base_dir):
            for sub_dir in sub_dirs[:]:
                if sub_dir.startswith('.') or \
                        sub_dir == '__pycache__' or \
                        os.path.join(root, sub_dir) in self.exclude_dirs:
                    sub_dirs.remove(sub_dir)

            for filename in files:
                if os.path.join(root, filename) not in self.exclude_files:
                    template_file_path = os.path.join(root, filename)
                    template_files.append(template_file_path)
        return template_files

    def render(self, **options):
        """Generate the network files."""
        source_files = []
        template_files = self.get_template_files()
        extensions = tuple([new_extension for _, new_extension in self.replace_template_suffixes])
        for template_file in template_files:
            new_file_path = template_file
            template_file_path = template_file
            for template_suffix, new_file_suffix in self.replace_template_suffixes:
                if new_file_path.endswith(template_suffix):
                    new_file_path = new_file_path[:-len(template_suffix)] + new_file_suffix
                    break

            source_file = SourceFile()
            source_file.file_relative_path = new_file_path[len(self.template_base_dir) + 1:]
            source_file.template_file_path = template_file_path
            if new_file_path.endswith(extensions):
                source_file.content = render_template(template_file_path, options)
            else:
                with open(template_file_path, encoding='utf-8') as fp:
                    source_file.content = fp.read()
            source_files.append(source_file)
        return source_files



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Common explain data encapsulator base class."""

import copy

from mindinsight.explainer.common.enums import ExplanationKeys
from mindinsight.utils.exceptions import ParamValueError


def _sort_key_min_confidence(sample, labels):
    """Samples sort key by the minimum confidence."""
    min_confidence = float("+inf")
    for inference in sample["inferences"]:
        if labels and inference["label"] not in labels:
            continue
        if inference["confidence"] < min_confidence:
            min_confidence = inference["confidence"]
    return min_confidence


def _sort_key_max_confidence(sample, labels):
    """Samples sort key by the maximum confidence."""
    max_confidence = float("-inf")
    for inference in sample["inferences"]:
        if labels and inference["label"] not in labels:
            continue
        if inference["confidence"] > max_confidence:
            max_confidence = inference["confidence"]
    return max_confidence


def _sort_key_min_confidence_sd(sample, labels):
    """Samples sort key by the minimum confidence_sd."""
    min_confidence_sd = float("+inf")
    for inference in sample["inferences"]:
        if labels and inference["label"] not in labels:
            continue
        confidence_sd = inference.get("confidence_sd", float("+inf"))
        if confidence_sd < min_confidence_sd:
            min_confidence_sd = confidence_sd
    return min_confidence_sd


def _sort_key_max_confidence_sd(sample, labels):
    """Samples sort key by the maximum confidence_sd."""
    max_confidence_sd = float("-inf")
    for inference in sample["inferences"]:
        if labels and inference["label"] not in labels:
            continue
        confidence_sd = inference.get("confidence_sd", float("-inf"))
        if confidence_sd > max_confidence_sd:
            max_confidence_sd = confidence_sd
    return max_confidence_sd


class ExplainDataEncap:
    """Explain data encapsulator base class."""

    def __init__(self, job_manager):
        self._job_manager = job_manager

    @property
    def job_manager(self):
        return self._job_manager


class ExplanationEncap(ExplainDataEncap):
    """Base encapsulator for explanation queries."""

    def __init__(self, image_url_formatter, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._image_url_formatter = image_url_formatter

    def _query_samples(self,
                       job,
                       labels,
                       sorted_name,
                       sorted_type,
                       prediction_types=None,
                       drop_type=None):
        """
        Query samples.

        Args:
            job (ExplainManager): Explain job to be query from.
            labels (list[str]): Label filter.
            sorted_name (str): Field to be sorted.
            sorted_type (str): Sorting order, 'ascending' or 'descending'.
            prediction_types (list[str]): Prediction type filter.
            drop_type (str, None): When it is None, all data will be kept. When it is 'hoc_layers', samples without
                hoc explanations will be drop out. When it is 'saliency_maps', samples without saliency explanations
                will be drop out.

        Returns:
             list[dict], samples to be queried.
        """

        samples = copy.deepcopy(job.get_all_samples())
        if drop_type not in (None, ExplanationKeys.SALIENCY.value, ExplanationKeys.HOC.value):
            raise ParamValueError(
                f"Argument drop_type valid options: None, {ExplanationKeys.SALIENCY.value}, "
                f"{ExplanationKeys.HOC.value}, but got {drop_type}.")

        if drop_type is not None:
            samples = [sample for sample in samples if any(infer[drop_type] for infer in sample['inferences'])]
        if labels:
            filtered = []
            for sample in samples:
                infer_labels = [inference["label"] for inference in sample["inferences"]]
                for infer_label in infer_labels:
                    if infer_label in labels:
                        filtered.append(sample)
                        break
            samples = filtered

        if prediction_types and len(prediction_types) < 3:
            filtered = []
            for sample in samples:
                infer_types = [inference["prediction_type"] for inference in sample["inferences"]]
                for infer_type in infer_types:
                    if infer_type in prediction_types:
                        filtered.append(sample)
                        break
            samples = filtered

        reverse = sorted_type == "descending"
        if sorted_name == "confidence":
            if reverse:
                samples.sort(key=lambda x: _sort_key_max_confidence(x, labels), reverse=reverse)
            else:
                samples.sort(key=lambda x: _sort_key_min_confidence(x, labels), reverse=reverse)
        elif sorted_name == "uncertainty":
            if not job.uncertainty_enabled:
                raise ParamValueError("Uncertainty is not enabled, sorted_name cannot be 'uncertainty'")
            if reverse:
                samples.sort(key=lambda x: _sort_key_max_confidence_sd(x, labels), reverse=reverse)
            else:
                samples.sort(key=lambda x: _sort_key_min_confidence_sd(x, labels), reverse=reverse)
        elif sorted_name != "":
            raise ParamValueError("sorted_name")
        return samples

    def _get_image_url(self, train_id, image_path, image_type):
        """Returns image's url."""
        if self._image_url_formatter is None:
            return image_path
        return self._image_url_formatter(train_id, image_path, image_type)



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The Parallel Data Analyser."""
import os
import re

from mindinsight.profiler.analyser.base_analyser import BaseAnalyser
from mindinsight.profiler.analyser.analyser_factory import AnalyserFactory


class ParallelAnalyser(BaseAnalyser):
    """Analyse parallel data from file."""

    _cluster_filename_reg = r'ascend_cluster_analyse_[a-z]+-parallel_[0-9]+_[0-9]+_[0-9]+\.csv'
    _step_trace_filename = 'step_trace_raw_{}_detail_time.csv'
    _flops_summary_filename = 'flops_summary_{}.json'
    _hccl_raw_filename = 'hccl_raw_{}.csv'
    _ascend_timeline_display_filename = 'ascend_timeline_display_{}.json'
    _cluster_filename = r'ascend_cluster_analyse_{}_{}_{}_{}.csv'

    _PIPELINE_PARALLEL = 'pipeline-parallel'

    _parallel_mode = ''
    _stage_num = 0
    _rank_size = 0
    _default_step = 0
    _cluster_step_trace_summary = []
    _flops_summary = []
    _rank_link_summary = []

    _RATE_DECIMAL_PLACES = 4

    def __init__(self, profiling_dir, rank_id):
        super().__init__(profiling_dir, rank_id)

        files = os.listdir(self._profiling_dir)
        for file in files:
            if not re.match(self._cluster_filename_reg, file):
                continue

            file_parts = file.split('_')
            self._parallel_mode = file_parts[3]
            self._stage_num = int(file_parts[4])
            self._rank_size = int(file_parts[5])
            break

    def _load(self):
        """Load data according to the parsed profiling files."""

    def _filter(self, filter_condition):
        """
        Filter the profiling data according to the filter condition.

        Args:
            filter_condition (dict): The filter condition.
        """

    def _set_cluster_step_trace_summary(self):
        """
        Get step interval summary.

        Returns:
            dict,the cluster step trace summary
            if parallel_mode =='data-parallel':['step_interval', 'computation_time', 'communication_alone_time']
            elif parallel_mode =='model-parallel':['step_interval','computation_time', 'communication_alone_time']
            elif parallel_mode =='pipeline-parallel':['step_interval','computation_time', 'communication_alone_time'
                , 'stage_time','receive_alone_time', 'collective_communication_alone_time']
        """
        self._cluster_step_trace_summary = []
        for rank_id in range(self._rank_size):
            cluster_step_trace_analyser = AnalyserFactory.instance().get_analyser(
                'cluster_step_trace', self._profiling_dir, self._device_id)
            cluster_step_trace = cluster_step_trace_analyser.get_step_bottleneck_info(rank_id, self._default_step)
            if cluster_step_trace:
                self._cluster_step_trace_summary.append({
                    'rank_id': rank_id,
                    'data': cluster_step_trace
                })

    def get_step_interval(self):
        """
        Get step interval summary.

        Returns:
            list, the step interval of ranks,list({"rank_id":0, "step_interval":23.45})
        """
        step_interval_summary = []
        if not self._cluster_step_trace_summary:
            self._set_cluster_step_trace_summary()
        for step_trace in self._cluster_step_trace_summary:
            step_interval_summary.append({
                "rank_id": step_trace["rank_id"],
                "step_interval": step_trace["data"][0]
            })
        return step_interval_summary

    def _set_flops_summary(self):
        """Get the flops of rank."""
        cluster_flops_analyser = AnalyserFactory.instance().get_analyser(
            'cluster_flops', self._profiling_dir, self._device_id)
        self._flops_summary = cluster_flops_analyser.get_flops()

    def get_rank_flops_per_second(self):
        """
        Get the number of floating point operations per second (the unit is billion).

        Returns:
            list, the step interval of ranks,list({"rank_id":0, "data":23.45})
        """
        if not self._flops_summary:
            self._set_flops_summary()

        flops = []
        for flop in self._flops_summary:
            flops.append({
                "rank_id": flop["rank_id"],
                "FLOPS": flop["FLOPS"] if flop.get("FLOPS") else flop["cube_FLOPS"] + flop["vec_FLOPS"]
            })

        return flops

    def get_rank_link_summary(self):
        """
        Get the rank link data.

        Returns:
            list, the rank link data,list([src_des, link_type
            , communication_cost, communication_size, communication_brand_width])
        """
        if not self._rank_link_summary:
            cluster_hccl_analyser = AnalyserFactory.instance().get_analyser(
                'cluster_hccl', self._profiling_dir, self._device_id)
            self._rank_link_summary = cluster_hccl_analyser.get_cluster_link_info()["cluster_link_info"]
        return self._rank_link_summary

    def get_communication_rate(self):
        """Get the proportion of communication time in the entire step."""
        computation_time = 0
        communication_time = 0
        communication_rate = 0.0

        if not self._cluster_step_trace_summary:
            self._set_cluster_step_trace_summary()

        computation_time = sum([step_trace["data"][1] for step_trace in self._cluster_step_trace_summary])
        communication_time = sum([step_trace["data"][2] for step_trace in self._cluster_step_trace_summary])

        if computation_time > 0 and communication_time > 0:
            communication_rate = round(communication_time / (computation_time + communication_time)
                                       , self._RATE_DECIMAL_PLACES)

        return communication_rate

    def get_between_stage_communication_rate(self):
        """Get the proportion of communication time between stages, here use the receive operator time."""
        computation_time = 0
        communication_time = 0
        receive_time = 0
        communication_rate = 0.0
        if not self._cluster_step_trace_summary:
            self._set_cluster_step_trace_summary()

        if self._parallel_mode == self._PIPELINE_PARALLEL:
            computation_time = sum([step_trace["data"][1] for step_trace in self._cluster_step_trace_summary])
            communication_time = sum([step_trace["data"][2] for step_trace in self._cluster_step_trace_summary])
            receive_time = sum([step_trace["data"][4] for step_trace in self._cluster_step_trace_summary])

        if computation_time > 0 and communication_time > 0 and receive_time > 0:
            communication_rate = round(receive_time / (computation_time + communication_time)
                                       , self._RATE_DECIMAL_PLACES)

        return communication_rate

    def get_within_stage_communication_rate(self):
        """
        Get the proportion of communication time within the stage
        , Here is the communication time after excluding the receive operator time
        """

        communication_rates = []
        if self._parallel_mode != self._PIPELINE_PARALLEL:
            return communication_rates

        if not self._cluster_step_trace_summary:
            self._set_cluster_step_trace_summary()

        stage_summary = {}
        without_receive_summary = {}
        for step_trace in self._cluster_step_trace_summary:
            stage_id = int(step_trace["rank_id"] // (self._rank_size / self._stage_num) + 1)
            without_receive_summary[stage_id] = without_receive_summary.get(stage_id, 0) + step_trace["data"][5]
            stage_summary[stage_id] = stage_summary.get(stage_id, 0) + step_trace["data"][1] + step_trace["data"][2]
        for stage_id in stage_summary:
            communication_rates.append({
                'stage_id': stage_id,
                'rate': without_receive_summary[stage_id] / stage_summary[stage_id]
            })

        return communication_rates

    def get_rank_flops(self):
        """Get all flops of rank."""
        if not self._flops_summary:
            self._set_flops_summary()

        flops = []
        for flop in self._flops_summary:
            flops.append({
                "rank_id": flop["rank_id"],
                "FLOPs": flop["FLOPs"] if flop.get("FLOPs") else flop["cube_FLOPs"] + flop["vec_FLOPs"]
            })

        return flops

    @property
    def parallel_mode(self):
        """Get parallel model."""
        return self._parallel_mode

    @property
    def stage_num(self):
        """Get stage num."""
        return self._stage_num

    @property
    def rank_size(self):
        """Get rank size."""
        return self._rank_size



# Copyright 2019-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Enums."""

import enum

class BaseEnum(enum.Enum):

    @classmethod
    def list_members(cls):
        """List all members."""
        return [member.value for member in cls]


class DataManagerStatus(BaseEnum):
    """Data manager status."""
    INIT = 'INIT'
    LOADING = 'LOADING'
    DONE = 'DONE'
    INVALID = 'INVALID'


class BriefCacheStatus(BaseEnum):
    """Brief cache status."""
    INIT = 'init'
    LOADING = 'loading'
    LOADED = 'loaded'


class PluginNameEnum(BaseEnum):
    """Plugin Name Enum."""
    IMAGE = 'image'
    SCALAR = 'scalar'
    GRAPH = 'graph'
    OPTIMIZED_GRAPH = 'optimized_graph'
    HISTOGRAM = 'histogram'
    TENSOR = 'tensor'
    LANDSCAPE = 'loss_landscape'


@enum.unique
class CacheStatus(enum.Enum):
    """Train job cache status."""
    NOT_IN_CACHE = "NOT_IN_CACHE"
    CACHING = "CACHING"
    CACHED = "CACHED"



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Common Tools."""
import imghdr
import math
import os

from numbers import Number

from mindinsight.datavisual.common.exceptions import MaxCountExceededError
from mindinsight.datavisual.common.exceptions import PathNotDirectoryError
from mindinsight.datavisual.common.log import logger
from mindinsight.utils import exceptions
from mindinsight.utils.exceptions import UnknownError

_IMG_EXT_TO_MIMETYPE = {
    'bmp': 'image/bmp',
    'gif': 'image/gif',
    'jpeg': 'image/jpeg',
    'png': 'image/png',
}
_DEFAULT_IMAGE_MIMETYPE = 'application/octet-stream'


def find_app_package():
    """Find package in current directory."""
    backend_dir = os.path.realpath(os.path.join(__file__, os.pardir, os.pardir, os.pardir, "backend"))
    packages = []
    for file in os.listdir(backend_dir):
        file_path = os.path.join(backend_dir, file)
        if os.path.isfile(file_path):
            continue
        if not os.path.isfile(os.path.join(file_path, '__init__.py')):
            continue
        rel_path = os.path.relpath(file_path, backend_dir)
        package = rel_path.replace(os.path.sep, '.')
        package = f"mindinsight.backend.{package}"
        packages.append(package)
    return packages


def to_str(bytes_or_text, encode="utf-8"):
    """Bytes transform string."""
    if isinstance(bytes_or_text, bytes):
        return bytes_or_text.decode(encode)
    if isinstance(bytes_or_text, str):
        return bytes_or_text

    raise TypeError("Param isn't str or bytes type, param={}".format(bytes_or_text))


def to_int(param, param_name):
    """
    Transfer param to int type.

    Args:
        param (Any): A param transformed.
        param_name (str): Param name.

    Returns:
        int, value after transformed.

    """
    try:
        param = int(param)
    except ValueError:
        raise exceptions.ParamTypeError(param_name, 'Integer')
    return param


def to_float(param, param_name):
    """
    Transfer param to float type.

    Args:
        param (Any): A param transformed.
        param_name (str): Param name.

    Returns:
        float, value after transformed.

    """
    try:
        param = float(param)
    except ValueError:
        raise exceptions.ParamTypeError(param_name, 'Float')
    return param


def str_to_bool(param, param_name):
    """
    Check param and transform it to bool.

    Args:
        param (str): 'true' or 'false' is valid.
        param_name (str): Param name.

    Returns:
        bool, if param is 'true', case insensitive.

    Raises:
        ParamValueError: If the value of param is not 'false' and 'true'.

    """
    if not isinstance(param, str):
        raise exceptions.ParamTypeError(param_name, 'str')

    if param.lower() not in ['false', 'true']:
        raise exceptions.ParamValueError("The value of %s must be 'false' or 'true'." % param_name)
    param = (param.lower() == 'true')

    return param


def get_img_mimetype(img_data):
    """
    Recognize image headers and generate image MIMETYPE.

    Args:
        img_data (bin): Binary character stream of image.

    Returns:
        str, a MIMETYPE of the give image.
    """
    image_type = imghdr.what(None, img_data)
    mimetype = _IMG_EXT_TO_MIMETYPE.get(image_type, _DEFAULT_IMAGE_MIMETYPE)
    return mimetype


def if_nan_inf_to_none(name, value):
    """
    Transform value to None if it is NaN or Inf.

    Args:
        name (str): Name of value.
        value (float): A number transformed.

    Returns:
        float, if value is NaN or Inf, return None.

    """
    if not isinstance(value, Number):
        raise exceptions.ParamTypeError(name, 'number')
    if math.isnan(value) or math.isinf(value):
        value = None
    return value


def exception_wrapper(func):
    """Exception wrapper"""
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (PathNotDirectoryError, FileNotFoundError) as err:
            # except PathNotDirectoryError and FileNotFoundError as they are on warning level
            logger.warning(str(err))
        except Exception as exc:
            logger.exception(exc)
            raise UnknownError(str(exc))
    return wrapper


def exception_no_raise_wrapper(func):
    """Don't raise exception to avoid printing error in stdout and log error in the log file."""
    def wrapper(*args, **kwargs):
        try:
            return exception_wrapper(func)(*args, **kwargs)
        except UnknownError as err:
            logger.error(str(err))
    return wrapper

class Counter:
    """Count accumulator with limit checking."""

    def __init__(self, max_count=None, init_count=0):
        self._count = init_count
        self._max_count = max_count

    def add(self, value=1):
        """Add value."""
        if self._max_count is not None and self._count + value > self._max_count:
            raise MaxCountExceededError()
        self._count += value



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Hook module."""

import os
import stat
import threading
from importlib import import_module

from mindinsight.conf import settings
from mindinsight.utils.exceptions import FileSystemPermissionError


class BaseHook:
    """Base hook class."""

    def register_secure_domains(self):
        """Hook function to register secure domains."""
        return []

    def register_startup_arguments(self, parser):
        """
        Hook function to register startup arguments.

        Args:
            parser (ArgumentParser): specify parser to which arguments are added.
        """

    def on_startup(self, logger):
        """
        Hook function to on startup.

        Args:
            logger (Logger): script logger of start command.
        """

    def on_shutdown(self, logger):
        """
        Hook function to on shutdown.

        Args:
            logger (Logger): script logger of stop command.
        """

    def on_init(self):
        """Hook function to on init."""


class HookUtils:
    """
    Lock utilities.

    Examples:
        >>> from mindinsight.utils.hook import HookUtils
        >>> for hook in HookUtils.instance().hooks():
        >>>     domains = hook.register_secure_domains()
        >>>     hook.register_startup_arguments(parser)
        >>>     hook.on_startup(logger)
        >>>     hook.on_shutdown(logger)
        >>>     hook.on_init()
    """

    _lock = threading.Lock()
    _instance = None

    def __new__(cls, *args, **kwargs):
        """Built-in __new__ function."""
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super().__new__(cls, *args, **kwargs)
                    cls._instance.discover()
        return cls._instance

    def discover(self):
        """Discover hook instances."""
        self.__hooks = []
        mindinsight_path = os.path.join(__file__, os.pardir, os.pardir)
        hook_path = os.path.realpath(os.path.join(mindinsight_path, 'common', 'hook'))
        files = os.listdir(hook_path)
        files.sort()
        for file in files:
            if file.startswith('_') or not file.endswith('.py'):
                continue
            hook_name = file[:-len('.py')]
            hook_module = import_module('mindinsight.common.hook.{}'.format(hook_name))
            hook_cls = getattr(hook_module, 'Hook', None)
            if hook_cls is not None and issubclass(hook_cls, BaseHook):
                self.__hooks.append(hook_cls())

    def hooks(self):
        """
        Return list of hook instances.

        Returns:
            list, list of hook instances.
        """
        return self.__hooks

    @classmethod
    def instance(cls):
        """
        Produce singleton instance of HookUtils.

        Returns:
            HookUtils, singleton instance of HookUtils.
        """
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance


def init(workspace='', config='', **kwargs):
    """
    Init MindInsight context.

    Args:
        workspace (str): specify mindinsight workspace, default is ''.
        config (str): specify mindinsight config file, default is ''.

    Raises:
        FileSystemPermissionError, if workspace is not allowed to access or available.
    """
    # set umask to 0o077
    os.umask(stat.S_IRWXG | stat.S_IRWXO)

    # assign argument values into environment
    if workspace:
        kwargs['workspace'] = workspace

    if config:
        kwargs['config'] = config

    for key, value in kwargs.items():
        variable = 'MINDINSIGHT_{}'.format(key.upper())
        os.environ[variable] = str(value)

    settings.refresh()

    if os.path.exists(settings.WORKSPACE):
        if not os.access(settings.WORKSPACE, os.R_OK | os.W_OK | os.X_OK):
            raise FileSystemPermissionError('Workspace {} not allowed to access'.format(workspace))
    else:
        try:
            os.makedirs(settings.WORKSPACE, mode=stat.S_IRWXU, exist_ok=True)
        except OSError:
            # race condition or priority problem
            raise FileSystemPermissionError('Workspace {} not available'.format(workspace))

    for hook in HookUtils.instance().hooks():
        hook.on_init()



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Define the watchpoint stream."""

import copy

from mindinsight.debugger.common.exceptions.exceptions import DebuggerParamValueError
from mindinsight.debugger.common.log import LOGGER as log
from mindinsight.debugger.common.utils import is_scope_type, is_cst_type
from mindinsight.debugger.conditionmgr.conditionmgr import ConditionMgr
from mindinsight.debugger.conditionmgr.common.utils import NodeBasicInfo
from mindinsight.debugger.conditionmgr.condition import ConditionIdEnum
from mindinsight.debugger.proto.debug_grpc_pb2 import WatchCondition

WATCHPOINT_CONDITION_MAPPING = {
    ConditionIdEnum.ACTIVATION_RANGE.value: WatchCondition.Condition.tensor_range,
    ConditionIdEnum.GRADIENT_EXPLODING.value: WatchCondition.Condition.tensor_general_overflow,
    ConditionIdEnum.GRADIENT_TOO_LARGE.value: WatchCondition.Condition.tensor_too_large,
    ConditionIdEnum.GRADIENT_VANISHING.value: WatchCondition.Condition.tensor_too_small,
    ConditionIdEnum.OPERATOR_OVERFLOW.value: WatchCondition.Condition.overflow,
    ConditionIdEnum.TENSOR_ALL_ZERO.value: WatchCondition.Condition.tensor_all_zero,
    ConditionIdEnum.TENSOR_OVERFLOW.value: WatchCondition.Condition.tensor_general_overflow,
    ConditionIdEnum.TENSOR_RANGE.value: WatchCondition.Condition.tensor_range,
    ConditionIdEnum.TENSOR_TOO_LARGE.value: WatchCondition.Condition.tensor_too_large,
    ConditionIdEnum.TENSOR_TOO_SMALL.value: WatchCondition.Condition.tensor_too_small,
    ConditionIdEnum.WEIGHT_CHANGE_TOO_LARGE.value: WatchCondition.Condition.tensor_change_too_large,
    ConditionIdEnum.WEIGHT_CHANGE_TOO_SMALL.value: WatchCondition.Condition.tensor_change_too_small,
    ConditionIdEnum.WEIGHT_INITIALIZATION.value: WatchCondition.Condition.tensor_initialization,
    ConditionIdEnum.WEIGHT_NOT_CHANGED.value: WatchCondition.Condition.tensor_not_changed,
    ConditionIdEnum.WEIGHT_OVERFLOW.value: WatchCondition.Condition.tensor_general_overflow,
    ConditionIdEnum.WEIGHT_TOO_LARGE.value: WatchCondition.Condition.tensor_too_large,
    ConditionIdEnum.WEIGHT_TOO_SMALL.value: WatchCondition.Condition.tensor_too_small
}


class WatchNodeTree:
    """The WatchNode Node Structure."""
    INVALID = -1  # the scope node and the nodes below are invalid
    NOT_WATCH = 0  # the scope node and the nodes below are not watched
    PARTIAL_WATCH = 1  # at least one node under the scope node is not watched
    TOTAL_WATCH = 2  # the scope node and the nodes below are all watched

    def __init__(self, node_name='', node_type=None, full_name='', watch_status=1):
        self._node_name = node_name
        self._full_name = full_name
        self._node_type = self._translate_node_type(node_type)
        self._watch_status = watch_status
        self._children = {}

    @property
    def node_name(self):
        """The property of node name."""
        return self._node_name

    @property
    def full_name(self):
        """The property of node name."""
        return self._full_name

    @property
    def node_type(self):
        """The property of node type."""
        return self._node_type

    @node_type.setter
    def node_type(self, value):
        """Set the node type."""
        self._node_type = self._translate_node_type(value)

    @property
    def watch_status(self):
        """The property of watch status about current node."""
        return self._watch_status

    @watch_status.setter
    def watch_status(self, value):
        """Set the node watch_status."""
        self._watch_status = value

    def update_metadata(self, node_type, full_name, watch_status):
        """Update the metadata for watched node."""
        self._full_name = full_name
        self._node_type = self._translate_node_type(node_type)
        self._watch_status = watch_status

    @staticmethod
    def _translate_node_type(node_type):
        """Translate node type to watch node type."""
        flag = node_type
        if not node_type or is_scope_type(node_type):
            flag = 'scope'
        return flag

    def get(self, sub_name):
        """Get sub node."""
        return self._children.get(sub_name)

    def get_children(self):
        """Get all children."""
        for name_scope, sub_watch_node in self._children.items():
            yield name_scope, sub_watch_node

    def get_children_count(self):
        """Get the count of children nodes."""
        return len(self._children)

    def add_node(self, node_name, node_type, full_name=''):
        """
        Add watch node to watch node tree.

        Args:
            node_name (str): The node name.
            node_type (str): The node type.
            full_name (str): The full name of node.
        """
        log.debug("Add node %s with type: %s, full_name: %s", node_name, node_type, full_name)
        scope_names = node_name.split('/', 1)
        if len(scope_names) == 1:
            target_node = self.get(node_name)
            if not target_node:
                self.add(node_name, node_type, full_name, watch_status=WatchNodeTree.TOTAL_WATCH)
            else:
                target_node.update_metadata(node_type, full_name, WatchNodeTree.TOTAL_WATCH)
            return

        scope_name, sub_names = scope_names
        sub_tree = self.get(scope_name)
        if not sub_tree:
            sub_tree = self.add(scope_name, watch_status=1)
        sub_tree.add_node(sub_names, node_type, full_name)

    def add(self, name, node_type=None, full_name='', watch_status=1):
        """Add sub WatchPointTree."""
        sub_name = '/'.join([self._node_name, name]) if self._node_name else name
        sub_tree = WatchNodeTree(sub_name, node_type, full_name, watch_status)
        self._children[name] = sub_tree

        return sub_tree

    def remove_node(self, node_name):
        """Remove sub node from current tree."""
        log.debug("Remove %s", node_name)
        scope_names = node_name.split('/', 1)
        sub_tree_name = scope_names[0]
        sub_tree = self._children.get(sub_tree_name)
        if not sub_tree:
            log.error("Failed to find node %s in WatchNodeTree.", sub_tree_name)
            raise DebuggerParamValueError("Failed to find node {}".format(sub_tree_name))

        if len(scope_names) > 1:
            sub_tree.remove_node(scope_names[1])

        if sub_tree.watch_status == WatchNodeTree.NOT_WATCH or len(scope_names) == 1:
            self._children.pop(sub_tree_name)

        self._watch_status = WatchNodeTree.PARTIAL_WATCH if self._children else \
            WatchNodeTree.NOT_WATCH


class Watchpoint:
    """
    The class of watchpoint stream.

    Args:
        watchpoint_id (int): The id of Watchpoint.
        watch_condition (dict): The condition of Watchpoint.

            - condition (str): Accept `INF` or `NAN`.
            - param (list[float]): Not defined yet.
    """

    def __init__(self, watchpoint_id, watch_condition, name=None):
        self._id = watchpoint_id
        self._condition = watch_condition
        self._watch_node = {0: WatchNodeTree()}
        self.name = name

    @property
    def watchpoint_id(self):
        """The property of watchpoint id."""
        return self._id

    @property
    def nodes(self):
        """The property of watch nodes."""
        return self._watch_node

    @property
    def condition(self):
        """The property of watch condition."""
        return self._condition

    def copy_nodes_from(self, other_watchpoint, deep_copy=False):
        """
        Copy nodes from other watchpoint.
        Args:
            other_watchpoint (Watchpoint): Other watchpoint.
            deep_copy (bool): Whether using deepcopy.
        """
        if deep_copy:
            self._watch_node = copy.deepcopy(other_watchpoint.nodes)
        else:
            self._watch_node = other_watchpoint.nodes

    def add_nodes(self, nodes, rank_id):
        """Add node into watchpoint."""
        if not nodes:
            log.warning("Add empty nodes.")
            return
        if rank_id not in self._watch_node:
            self._watch_node[rank_id] = WatchNodeTree()
        if not isinstance(nodes, list):
            nodes = [nodes]
        for node in nodes:
            watch_node = self._watch_node.get(rank_id)
            watch_node.add_node(node.name, node.type, node.full_name)

    def remove_nodes(self, nodes, rank_id):
        """Remove nodes from watchpoint."""
        if not nodes:
            return
        self.validate_rank_id(rank_id)
        if not isinstance(nodes, list):
            nodes = [nodes]
        for node in nodes:
            self._watch_node.get(rank_id).remove_node(node.name)

    def get_node_status(self, node_name, node_type, full_name, rank_id):
        """Judge if the node is in watch nodes."""
        if is_cst_type(node_type):
            return WatchNodeTree.INVALID
        scope_names = node_name.split('/')
        self.validate_rank_id(rank_id)
        cur_node = self._watch_node.get(rank_id)
        status = 1
        for scope_name in scope_names:
            cur_node = cur_node.get(scope_name)
            if cur_node is None:
                status = WatchNodeTree.NOT_WATCH
                break
            if cur_node.watch_status == WatchNodeTree.TOTAL_WATCH:
                status = WatchNodeTree.TOTAL_WATCH
                break
        if status == WatchNodeTree.TOTAL_WATCH and cur_node.node_name != node_name:
            self._watch_node.get(rank_id).add_node(node_name, node_type, full_name)

        return status

    def _get_watch_node(self, cur_watch_node, watch_node_list):
        """
        Traverse the watch nodes and add total watched node list to `watch_node_list`.

        Args:
            cur_watch_node (WatchNodeTree): The current watch node.
            watch_node_list (list[NodeBasicInfo]): The list of watch node basic infos.
        """
        if cur_watch_node.watch_status == WatchNodeTree.TOTAL_WATCH:
            node_info = NodeBasicInfo(name=cur_watch_node.node_name,
                                      full_name=cur_watch_node.full_name,
                                      type=cur_watch_node.node_type)
            watch_node_list.append(node_info)
            return
        for _, watch_node in cur_watch_node.get_children():
            self._get_watch_node(watch_node, watch_node_list)

    def get_watch_nodes(self):
        """
        Get the name of all total watched nodes.

        Returns:
            list[NodeBasicInfo], the list of watch node basic infos.
        """
        watch_nodes_for_devices = {}
        for rank_id, watch_node_tree in self._watch_node.items():
            watch_nodes = []
            self._get_watch_node(watch_node_tree, watch_nodes)
            watch_nodes_for_devices[rank_id] = watch_nodes
        return watch_nodes_for_devices

    def add_set_cmd_detail(self, set_cmd, watch_nodes_for_devices):
        """Return the watchpoint in proto format."""
        # construct SetCMD
        condition_id = self._condition.get('id')
        set_cmd.id = self._id
        set_cmd.delete = False
        set_cmd.watch_condition.condition = WATCHPOINT_CONDITION_MAPPING.get(condition_id)
        condition_mgr = ConditionMgr()
        condition = condition_mgr.get_condition(condition_id)
        param_dict = {
            param.get('name'): param for param in self._condition.get('params')
        }
        for param_name in condition.ordered_parameter_names:
            param = param_dict.get(param_name)
            if param:
                param_proto = set_cmd.watch_condition.params.add()
                param_proto.name = param.get('name')
                param_proto.value = param.get('value')
                param_proto.disabled = False
                # Only one parameter of condition in old mindspore version.
                set_cmd.watch_condition.value = param.get('value')
            else:
                param_proto = set_cmd.watch_condition.params.add()
                param_proto.name = param_name
                param_proto.disabled = True

        for rank_id, watch_nodes in watch_nodes_for_devices.items():
            for watch_node in watch_nodes:
                event_node = set_cmd.watch_nodes.add()
                event_node.graph_name = watch_node.name.split('/', 1)[0]
                event_node.node_name = watch_node.full_name
                event_node.node_type = watch_node.type
                event_node.rank_id = rank_id

    def get_watch_condition_info(self):
        """Get watch condition info."""
        watchpoint_info = {
            'id': self._id,
            'watch_condition': self._condition
        }
        if self.name:
            watchpoint_info['name'] = self.name
        return watchpoint_info

    def validate_rank_id(self, rank_id):
        """Validate the rank id."""
        if rank_id not in self._watch_node:
            log.warning("Rank_id not exist")
            return


class WatchpointHit:
    """The watchpoint hit structure."""

    def __init__(self, tensor_proto, watchpoint, node_name, graph_name):
        self._full_name = tensor_proto.node_name
        self._watchpoint = watchpoint
        self.node_name = node_name
        self.slot = tensor_proto.slot
        self.graph_name = graph_name
        self.error_code = 0

    @property
    def tensor_full_name(self):
        """The property of tensor full name."""
        tensor_name = ':'.join([self._full_name, self.slot])
        return tensor_name

    @property
    def watchpoint(self):
        """The property of watchpoint."""
        watchpoint = self._watchpoint.get_watch_condition_info()
        return watchpoint

    def __eq__(self, other):
        """Define the equal condition."""
        flag = self.tensor_full_name == other.tensor_full_name \
               and self.watchpoint == other.watchpoint \
               and self.graph_name == other.graph_name
        return flag



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""This file is used to define the GraphHistory."""


class GraphHistory:
    """Graph history of a device."""

    def __init__(self, rank_id, graph_history):
        self._rank_id = rank_id
        # the key is graph id, the value is list of executed graph run id.
        self._graph_history = graph_history
        self._count = None

    @property
    def rank_id(self):
        """Return rank id."""
        return self._rank_id

    @property
    def history(self):
        """Return execution history of all graphs."""
        return self._graph_history

    def clear(self):
        """Clear Graph History."""
        self._graph_history = {}
        self._count = 0

    def get_total_count(self):
        """Get the last graph run id."""
        if self._count is not None:
            return self._count
        last_id = -1
        for history in self._graph_history.values():
            if not history:
                continue
            last_id = max(last_id, max(history))
        self._count = last_id + 1
        return self._count

    def get_run_ids(self, graph_id):
        """
        Get executed graph run ids.

        Args:
            graph_id (int): The root graph id.

        Returns:
            set[int], the list of graph run ids.
        """
        return self._graph_history.get(graph_id, set())



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Validate the input path."""
import os
import re
from typing import Union, List

from marshmallow import ValidationError

from mindinsight.profiler.common.exceptions.exceptions import \
    ProfilerParamValueErrorException, ProfilerDirNotFoundException
from mindinsight.datavisual.common.exceptions import TrainJobNotExistError
from mindinsight.profiler.common.log import logger as log


def safe_normalize_path(
        path,
        raise_key,
        safe_prefixes: Union[None, List[str]],
        check_absolute_path=False,
        allow_parent_dir=False,
):
    """
    Returns safe normalized path.

    This func validates given path, and returns its normalized form. If
    safe_prefixes is given, this func will check whether the path is safe.

    Note:
        This func is not compatible with windows.

        Caller should check returned path to ensure safety according to
        business logic.

        File scheme (rfc8089) is currently not supported.

    Args:
        path (str): Path to be normalized.

        raise_key (str): The exception raise key

        safe_prefixes (list[str]): If not none, path must startswith one of the
            safe_prefixes. Set this arg to [] will cause all paths considered
            unsafe. Normally, prefix in this arg should end with "/".

        check_absolute_path (bool): Whether check path is absolute.

        allow_parent_dir (bool): Whether allow parent dir in path.

    Returns:
        str, normalized path.
    """
    normalized_path = validate_and_normalize_path(
        path,
        raise_key=raise_key,
        check_absolute_path=check_absolute_path,
        allow_parent_dir=allow_parent_dir,
    )

    if safe_prefixes is None:
        return normalized_path

    normalized_str = str(normalized_path)
    for prefix in safe_prefixes:
        if normalized_str.startswith(prefix):
            return normalized_path

    raise ValidationError({raise_key: {"The path is invalid!"}})


def validate_and_normalize_path(
        path,
        raise_key,
        check_absolute_path=False,
        allow_parent_dir=False,
):
    """
    Validates path and returns its normalized form.

    If path has a valid scheme, treat path as url, otherwise consider path a
    unix local path.

    Note:
        File scheme (rfc8089) is currently not supported.

    Args:
        path (str): Path to be normalized.
        raise_key (str): The exception raise key.
        check_absolute_path (bool): Whether check path scheme is supported.
        allow_parent_dir (bool): Whether allow parent dir in path.

    Returns:
        str, normalized path.
    """
    if not path:
        raise ValidationError({raise_key: {"The path is invalid!"}})

    path_str = str(path)
    if not allow_parent_dir:
        path_components = path_str.split("/")
        if ".." in path_components:
            raise ValidationError({raise_key: {"The path is invalid!"}})

    # path does not have valid schema, treat it as unix local path.
    if check_absolute_path:
        if not path_str.startswith("/"):
            raise ValidationError({raise_key: {"The path is invalid!"}})
    try:
        # most unix systems allow
        normalized_path = os.path.realpath(path)
    except ValueError:
        raise ValidationError({raise_key: {"The path is invalid!"}})

    return normalized_path


def validate_and_normalize_profiler_path(summary_dir, summary_base_dir):
    """
    Validate and normalize profiler path.

    Args:
        summary_dir (str): The relative path of summary directory.
        summary_base_dir (str): The summary base directory.

    Returns:
        str, normalized path of profiler directory.
    """
    profiler_directory_pattern = r'^profiler.*'
    if not summary_dir:
        raise ProfilerParamValueErrorException('The file dir does not exist.')
    train_job_dir = os.path.join(summary_base_dir, summary_dir)
    try:
        train_job_dir_abs = validate_and_normalize_path(train_job_dir, 'train_job_dir')
    except ValidationError:
        log.error('train_job dir <%s> is invalid', train_job_dir)
        raise ProfilerParamValueErrorException('train_job dir is invalid.')
    if not os.path.exists(train_job_dir_abs):
        raise TrainJobNotExistError(error_detail=train_job_dir_abs)

    try:
        profiler_name_list = []
        for dir_name in os.listdir(train_job_dir_abs):
            search_res = re.search(profiler_directory_pattern, dir_name)
            if search_res:
                profiler_name_list.append(search_res[0])
        profiler_name_list.sort()
        profiler_name_newest = profiler_name_list[-1]
        profiler_dir = os.path.join(summary_base_dir, summary_dir, profiler_name_newest)
    except ValidationError:
        log.error('no valid profiler dir under <%s>', train_job_dir_abs)
        raise ProfilerDirNotFoundException('Profiler dir not found.')
    try:
        profiler_dir = validate_and_normalize_path(profiler_dir, 'profiler')
    except ValidationError:
        log.error('profiler dir <%s> is invalid', profiler_dir)
        raise ProfilerParamValueErrorException('Profiler dir is invalid.')

    return profiler_dir



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Enums."""

import enum


class BaseEnum(enum.Enum):
    """Base enum."""
    @classmethod
    def list_members(cls):
        """List all members."""
        return [member.value for member in cls]


class ReasonCode(BaseEnum):
    """Reason code for calculating importance."""
    NOT_ALL_NUMBERS = 1
    SAMPLES_NOT_ENOUGH = 2
    CORRELATION_NAN = 3


class AcquisitionFunctionEnum(BaseEnum):
    """Enum for acquisition function method."""
    # Upper confidence bound
    UCB = 'ucb'
    # Probability of improvement
    PI = 'pi'
    # Expected improvement
    EI = 'ei'


class TuneMethod(BaseEnum):
    """Enum for tuning method."""
    # Gaussian process regressor
    GP = 'gp'


class GPSupportArgs(BaseEnum):
    METHOD = 'method'


class HyperParamKey(BaseEnum):
    """Config keys for hyper parameters."""
    BOUND = 'bounds'
    CHOICE = 'choice'
    DECIMAL = 'decimal'
    TYPE = 'type'
    SOURCE = 'source'


class HyperParamType(BaseEnum):
    """Config keys for hyper parameters."""
    INT = 'int'
    FLOAT = 'float'


class TargetKey(BaseEnum):
    """Config keys for target."""
    GROUP = 'group'
    NAME = 'name'
    GOAL = 'goal'


class TargetGoal(BaseEnum):
    """Goal for target."""
    MAXIMUM = 'maximize'
    MINIMUM = 'minimize'


class HyperParamSource(BaseEnum):
    SYSTEM_DEFINED = 'system_defined'
    USER_DEFINED = 'user_defined'


class TargetGroup(BaseEnum):
    SYSTEM_DEFINED = 'system_defined'
    METRIC = 'metric'


class TunableSystemDefinedParams(BaseEnum):
    """Tunable metadata keys of lineage collection."""
    BATCH_SIZE = 'batch_size'
    EPOCH = 'epoch'
    LEARNING_RATE = 'learning_rate'


class SystemDefinedTargets(BaseEnum):
    """System defined targets"""
    LOSS = 'loss'



# Copyright 2020-2022 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Debugger error code and messages."""
from enum import Enum, unique
from mindinsight.utils.constant import DebuggerErrors as DebuggerErrorCodes


_PARAM_ERROR_MASK = 0b00001 << 7
_DEBUGGER_GRAPH_ERROR = 0b00010 << 7
_DEBUGGER_RUNNING_ERROR = 0b00011 << 7
_DEBUGGER_SERVER_ERROR = 0b00100 << 7
_DEBUGGER_SESSION_ERROR = 0b00101 << 7
_DEBUGGER_HISTORY_ERROR = 0b00110 << 7


@unique
class DebuggerErrors(DebuggerErrorCodes):
    """Debugger error codes."""
    PARAM_TYPE_ERROR = 0 | _PARAM_ERROR_MASK
    PARAM_VALUE_ERROR = 1 | _PARAM_ERROR_MASK
    STEP_NUM_ERROR = 2 | _PARAM_ERROR_MASK
    DEBUGGER_CONDITION_UNAVAILABLE_ERROR = 3 | _PARAM_ERROR_MASK

    NODE_NOT_IN_GRAPH_ERROR = 0 | _DEBUGGER_GRAPH_ERROR
    GRAPH_NOT_EXIST_ERROR = 1 | _DEBUGGER_GRAPH_ERROR
    NODE_TOO_LARGE_ERROR = 2 | _DEBUGGER_GRAPH_ERROR

    CREATE_WATCHPOINT_ERROR = 0 | _DEBUGGER_RUNNING_ERROR
    UPDATE_WATCHPOINT_ERROR = 1 | _DEBUGGER_RUNNING_ERROR
    DELETE_WATCHPOINT_ERROR = 2 | _DEBUGGER_RUNNING_ERROR
    CONTINUE_ERROR = 3 | _DEBUGGER_RUNNING_ERROR
    PAUSE_ERROR = 4 | _DEBUGGER_RUNNING_ERROR
    COMPARE_TENSOR_ERROR = 5 | _DEBUGGER_RUNNING_ERROR
    RECHECK_ERROR = 6 | _DEBUGGER_RUNNING_ERROR
    TENSOR_GRAPH_ERROR = 7 | _DEBUGGER_RUNNING_ERROR
    TENSOR_HIT_ERROR = 8 | _DEBUGGER_RUNNING_ERROR
    SET_RECOMMEND_WATCHPOINT_ERROR = 9 | _DEBUGGER_RUNNING_ERROR

    DEBUGGER_SERVER_RUNNING_ERROR = 0 | _DEBUGGER_SERVER_ERROR
    DEVICE_ID_UNREGISTERED = 1 | _DEBUGGER_SERVER_ERROR
    MODULE_NOT_FOUND_ERROR = 2 | _DEBUGGER_SERVER_ERROR
    TOOLKIT_NOT_FOUND_ERROR = 3 | _DEBUGGER_SERVER_ERROR

    DEBUGGER_SESSION_OVER_BOUND_ERROR = 0 | _DEBUGGER_SESSION_ERROR
    DEBUGGER_SESSION_NOT_FOUND_ERROR = 1 | _DEBUGGER_SESSION_ERROR
    DEBUGGER_ONLINE_SESSION_UNAVAILABLE = 2 | _DEBUGGER_SESSION_ERROR
    DEBUGGER_DOWNLOAD_OVER_QUEUE = 3 | _DEBUGGER_SESSION_ERROR
    DEBUGGER_DOWNLOAD_TENSOR_NOT_EXIST = 4 | _DEBUGGER_SESSION_ERROR
    DEBUGGER_RANK_DIR_NOT_FOUND = 5 | _DEBUGGER_SESSION_ERROR
    DEBUGGER_JSON_FILE_PARSE_ERROR = 6 | _DEBUGGER_SESSION_ERROR

    DEBUGGER_HISTORY_NOT_FOUND_ERROR = 0 | _DEBUGGER_HISTORY_ERROR
    DEBUGGER_STEP_VALUE_ERROR = 1 | _DEBUGGER_HISTORY_ERROR


@unique
class DebuggerErrorMsg(Enum):
    """Debugger error messages."""
    PARAM_TYPE_ERROR = "TypeError. {}"
    PARAM_VALUE_ERROR = "ValueError. {}"
    DEBUGGER_CONDITION_UNAVAILABLE_ERROR = "Condition is unavailable. {}"

    GRAPH_NOT_EXIST_ERROR = "The graph does not exist."
    NODE_TOO_LARGE_ERROR = "The nodes in dump directory is too large to analyze. " \
                           "Node limit: {}, current node num: {}."

    CREATE_WATCHPOINT_ERROR = "Create watchpoint failed. {}"
    UPDATE_WATCHPOINT_ERROR = "Update watchpoint failed. {}"
    DELETE_WATCHPOINT_ERROR = "Delete watchpoint failed. {}"
    CONTINUE_ERROR = "Continue debugging failed. {}"
    PAUSE_ERROR = "Pause debugging failed. {}"
    RECHECK_ERROR = "Recheck failed. {}"
    TENSOR_GRAPH_ERROR = "Get tensor graphs failed."
    TENSOR_HIT_ERROR = "Get tensor hits failed."
    SET_RECOMMEND_WATCHPOINT_ERROR = "Set Recommend Watchpoints failed."

    DEBUGGER_SERVER_RUNNING_ERROR = "Debugger server running error. {}"
    DEVICE_ID_UNREGISTERED = "Device id unregistered. Device id: {}"
    MODULE_NOT_FOUND_ERROR = "{} module not found."
    TOOLKIT_NOT_FOUND_ERROR = "File not found. {}"

    DEBUGGER_SESSION_OVER_BOUND_ERROR = "The amount of sessions is over limitation."
    DEBUGGER_SESSION_NOT_FOUND_ERROR = "Session {} not found."
    DEBUGGER_ONLINE_SESSION_UNAVAILABLE = "Online session is unavailable."
    DEBUGGER_DOWNLOAD_OVER_QUEUE = "There is already a tensor in download"
    DEBUGGER_DOWNLOAD_TENSOR_NOT_EXIST = "No such tensor to download"
    DEBUGGER_RANK_DIR_NOT_FOUND = "No rank directory found under {}."
    DEBUGGER_JSON_FILE_PARSE_ERROR = "Failed to parse the json files. {}"

    DEBUGGER_HISTORY_NOT_FOUND_ERROR = "The history of graph id: {} is not found."
    DEBUGGER_STEP_VALUE_ERROR = "The dumped step value of graph {} is not the subset of graph history."
